"""Module for processing impedance cardiography (ICG) signals."""

from biopsykit.signals.icg import event_extraction, outlier_correction, preprocessing

__all__ = ["event_extraction", "outlier_correction", "preprocessing"]



"""Module providing various functions for low-level handling of array data."""

from typing import Optional, Union

import neurokit2 as nk
import numpy as np
import pandas as pd
from scipy import interpolate, signal

from biopsykit.utils._datatype_validation_helper import _assert_num_columns
from biopsykit.utils._types_internal import arr_t, str_t


def sanitize_input_1d(data: arr_t) -> np.ndarray:
    """Convert 1-d array-like data (:class:`~pandas.DataFrame`/:class:`~pandas.Series`) to a numpy array.

    Parameters
    ----------
    data : array_like
        input data. Needs to be 1-d


    Returns
    -------
    :class:`~numpy.ndarray`
        data as 1-d :class:`~numpy.ndarray`

    """
    if isinstance(data, (pd.Series, pd.DataFrame)):
        # only 1-d pandas DataFrame allowed
        if isinstance(data, pd.DataFrame) and len(data.columns) != 1:
            raise ValueError("Only 1-d dataframes allowed!")
        data = np.squeeze(data.values)

    return data


def sanitize_input_nd(
    data: arr_t,
    ncols: Optional[Union[int, tuple[int, ...]]] = None,
) -> np.ndarray:
    """Convert n-d array-like data (:class:`~pandas.DataFrame`/:class:`~pandas.Series`) to a numpy array.

    Parameters
    ----------
    data : array_like
        input data
    ncols : int or tuple of ints
        number of columns (2nd dimension) the ``data`` is expected to have, a list of such if ``data``
        can have a set of possible column numbers or ``None`` to allow any number of columns. Default: ``None``


    Returns
    -------
    :class:`~numpy.ndarray`
        data as n-d numpy array

    """
    # ensure list
    if isinstance(ncols, int):
        ncols = [ncols]

    if isinstance(data, (pd.Series, pd.DataFrame)):
        data = data.to_numpy()

    if ncols is not None:
        if data.ndim == 1:
            if 1 in ncols:
                return data
            raise ValueError(f"Invalid number of columns! Expected one of {ncols}, got 1.")
        if data.shape[1] not in ncols:
            raise ValueError(f"Invalid number of columns! Expected one of {ncols}, got {data.shape[1]}.")
    return data


def sanitize_input_series(data: Union[np.ndarray, pd.Series, pd.DataFrame], name: str) -> pd.Series:
    """Convert input data to a pandas Series.

    Parameters
    ----------
    data : array_like
        input data
    name : str
        name of the series

    Returns
    -------
    :class:`~pandas.Series`
        input data as a pandas Series

    """
    if isinstance(data, pd.Series):
        return data
    if isinstance(data, pd.DataFrame):
        if len(data.columns) > 1:
            raise ValueError("Dataframe must have only one column to be converted to a Series!")
        return data.squeeze()
    return pd.Series(data, name=name)


def sanitize_input_dataframe_1d(data: Union[np.ndarray, pd.Series, pd.DataFrame], column: str) -> pd.DataFrame:
    """Convert input data to a pandas DataFrame with one column.

    Parameters
    ----------
    data : array_like
        input data
    columns : str
        name of the column

    Returns
    -------
    :class:`~pandas.DataFrame`
        input data as a pandas DataFrame

    """
    if isinstance(data, pd.DataFrame):
        _assert_num_columns(data, 1)
        return data.rename(columns={data.columns[0]: column})
    if isinstance(data, pd.Series):
        return data.to_frame(name=column)
    return pd.DataFrame(data, columns=[column])


def find_extrema_in_radius(
    data: arr_t,
    indices: arr_t,
    radius: Union[int, tuple[int, int]],
    extrema_type: Optional[str] = "min",
) -> np.ndarray:
    """Find extrema values (min or max) within a given radius around array indices.

    Parameters
    ----------
    data : array_like
        input data
    indices : array_like
        array with indices for which to search for extrema values around
    radius: int or tuple of int
        radius around ``indices`` to search for extrema:

        * if ``radius`` is an ``int`` then search for extrema equally in both directions in the interval
          [index - radius, index + radius].
        * if ``radius`` is a ``tuple`` then search for extrema in the interval
          [ index - radius[0], index + radius[1] ]

    extrema_type : {'min', 'max'}, optional
        extrema type to be searched for. Default: 'min'

    Returns
    -------
    :class:`~numpy.ndarray`
        numpy array containing the indices of the found extrema values in the given radius around ``indices``.
        Has the same length as ``indices``.

    Examples
    --------
    >>> from biopsykit.utils.array_handling import find_extrema_in_radius
    >>> data = pd.read_csv("data.csv")
    >>> indices = np.array([16, 25, 40, 57, 86, 100])
    >>>
    >>> radius = 4
    >>> # search minima in 'data' in a 4 sample 'radius' around each entry of 'indices'
    >>> find_extrema_in_radius(data, indices, radius)
    >>>
    >>> radius = (5, 0)
    >>> # search maxima in 'data' in a 5 samples before each entry of 'indices'
    >>> find_extrema_in_radius(data, indices, radius, extrema_type='max')

    """
    extrema_funcs = {"min": np.nanargmin, "max": np.nanargmax}

    if extrema_type not in extrema_funcs:
        raise ValueError(f"`extrema_type` must be one of {list(extrema_funcs.keys())}, not {extrema_type}")
    extrema_func = extrema_funcs[extrema_type]

    # ensure numpy
    data = sanitize_input_1d(data)
    indices = sanitize_input_1d(indices)
    indices = indices.astype(int)
    # possible start offset if beginning of array needs to be padded to ensure radius
    start_padding = 0

    lower_limit, upper_limit = _find_extrema_in_radius_get_limits(radius)

    # pad end/start of array if last_index+radius/first_index-radius is longer/shorter than array
    if len(data) - np.max(indices) <= upper_limit:
        data = np.pad(data, (0, upper_limit), constant_values=np.nan)
    if np.min(indices) < lower_limit:
        start_padding = lower_limit
        data = np.pad(data, (lower_limit, 0), constant_values=np.nan)

    # initialize window array
    windows = np.zeros(shape=(len(indices), lower_limit + upper_limit + 1))
    for i, index in enumerate(indices):
        # get windows around index
        windows[i] = data[index - lower_limit + start_padding : index + upper_limit + start_padding + 1]

    return extrema_func(windows, axis=1) + indices - lower_limit


def _find_extrema_in_radius_get_limits(radius: Union[int, tuple[int, int]]) -> tuple[int, int]:
    # determine upper and lower limit
    lower_limit = radius[0] if isinstance(radius, tuple) else radius
    upper_limit = radius[-1] if isinstance(radius, tuple) else radius

    # round up and make sure it's an integer
    lower_limit = np.ceil(lower_limit).astype(int)
    upper_limit = np.ceil(upper_limit).astype(int)
    return lower_limit, upper_limit


def remove_outlier_and_interpolate(
    data: arr_t,
    outlier_mask: np.ndarray,
    x_old: Optional[np.ndarray] = None,
    desired_length: Optional[int] = None,
) -> np.ndarray:
    """Remove outliers, impute missing values and optionally interpolate data to desired length.

    Detected outliers are removed from array and imputed by linear interpolation.
    Optionally, the output array can be linearly interpolated to a new length.


    Parameters
    ----------
    data : array_like
        input data
    outlier_mask : :class:`~numpy.ndarray`
        boolean outlier mask. Has to be the same length as ``data``. ``True`` entries indicate outliers.
        If ``outlier_mask`` is not a bool array values will be casted to bool
    x_old : array_like, optional
        x values of the input data to be interpolated or ``None`` if output data should not be interpolated
        to new length. Default: ``None``
    desired_length : int, optional
        desired length of the output signal or ``None`` to keep length of input signal. Default: ``None``


    Returns
    -------
    :class:`~numpy.ndarray`
        data with removed and imputed outliers, optionally interpolated to desired length


    Raises
    ------
    ValueError
        if ``data`` and ``outlier_mask`` don't have the same length or if ``x_old`` is ``None`` when ``desired_length``
        is passed as parameter

    """
    # ensure numpy and ensure that outlier_mask is a boolean array
    data = sanitize_input_nd(data)
    outlier_mask = np.array(outlier_mask).astype(bool)

    if len(data) != len(outlier_mask):
        raise ValueError("'data' and 'outlier_mask' need to have same length!")
    if x_old is None and desired_length:
        raise ValueError("'x_old' must also be passed when 'desired_length' is passed!")
    x_old = np.array(x_old)

    # remove outlier
    data[outlier_mask] = np.nan
    # fill outlier by linear interpolation of neighbors
    data = pd.Series(data).interpolate(limit_direction="both").to_numpy()
    # interpolate signal
    x_new = np.linspace(x_old[0], x_old[-1], desired_length)
    return nk.signal_interpolate(x_old, data, x_new, method="linear")


def sliding_window(
    data: arr_t,
    window_samples: Optional[int] = None,
    window_sec: Optional[int] = None,
    sampling_rate: Optional[float] = None,
    overlap_samples: Optional[int] = None,
    overlap_percent: Optional[float] = None,
) -> np.ndarray:
    """Create sliding windows from an input array.

    The window size of sliding windows can either be specified in *samples* (``window_samples``)
    or in *seconds* (``window_sec``, together with ``sampling_rate``).

    The overlap of windows can either be specified in *samples* (``overlap_samples``)
    or in *percent* (``overlap_percent``).

    .. note::
        If ``data`` has more than one dimension the sliding window view is applied to the **first** dimension.
        In the 2-d case this would correspond to applying windows along the **rows**.


    Parameters
    ----------
    data : array_like
        input data
    window_samples : int, optional
        window size in samples or ``None`` if window size is specified in seconds + sampling rate. Default: ``None``
    window_sec : int, optional
        window size in seconds or ``None`` if window size is specified in samples. Default: ``None``
    sampling_rate : float, optional
        sampling rate of data in Hz. Only needed if window size is specified in seconds (``window_sec`` parameter).
        Default: ``None``
    overlap_samples : int, optional
        overlap of windows in samples or ``None`` if window overlap is specified in percent. Default: ``None``
    overlap_percent : float, optional
        overlap of windows in percent or ``None`` if window overlap is specified in samples. Default: ``None``


    Returns
    -------
    :class:`~numpy.ndarray`
        sliding windows from input array.


    See Also
    --------
    :func:`~biopsykit.utils.array_handling.sliding_window_view`
        create sliding window of input array. low-level function with less input parameter configuration possibilities

    """
    # check input
    data = sanitize_input_nd(data)

    window, overlap = sanitize_sliding_window_input(
        window_samples=window_samples,
        window_sec=window_sec,
        sampling_rate=sampling_rate,
        overlap_samples=overlap_samples,
        overlap_percent=overlap_percent,
    )

    return sliding_window_view(data, window_length=window, overlap=overlap, nan_padding=True)


def sanitize_sliding_window_input(
    window_samples: Optional[int] = None,
    window_sec: Optional[int] = None,
    sampling_rate: Optional[float] = None,
    overlap_samples: Optional[int] = None,
    overlap_percent: Optional[float] = None,
) -> tuple[int, int]:
    """Sanitize input parameters for creating sliding windows from array data.

    The window size of sliding windows can either be specified in *samples* (``window_samples``)
    or in *seconds* (``window_sec``, together with ``sampling_rate``).

    The overlap of windows can either be specified in *samples* (``overlap_samples``)
    or in *percent* (``overlap_percent``).


    Parameters
    ----------
    window_samples : int, optional
        window size in samples or ``None`` if window size is specified in seconds + sampling rate. Default: ``None``
    window_sec : int, optional
        window size in seconds or ``None`` if window size is specified in samples. Default: ``None``
    sampling_rate : float, optional
        sampling rate of data in Hz. Only needed if window size is specified in seconds (``window_sec`` parameter).
        Default: ``None``
    overlap_samples : int, optional
        overlap of windows in samples or ``None`` if window overlap is specified in percent. Default: ``None``
    overlap_percent : float, optional
        overlap of windows in percent or ``None`` if window overlap is specified in samples. Default: ``None``


    Returns
    -------
    window : int
        window size in samples
    overlap : int
        window overlap in samples

    """
    if all(x is None for x in (window_samples, window_sec, sampling_rate)):
        raise ValueError(
            "Either 'window_samples', or 'window_sec' in combination with "
            "'sampling_rate' must be supplied as parameter!"
        )

    if window_samples is None:
        if sampling_rate == 0:
            raise ValueError("Sampling rate must be specified when 'window_sec' is used!")
        window = int(sampling_rate * window_sec)
    else:
        window = int(window_samples)

    overlap = _compute_overlap_samples(window, overlap_samples, overlap_percent)

    return window, overlap


def _compute_overlap_samples(
    window: int, overlap_samples: Optional[int] = None, overlap_percent: Optional[float] = None
):
    if overlap_samples is not None:
        overlap = int(overlap_samples)
    elif overlap_percent is not None:
        if overlap_percent > 1:
            overlap_percent /= 100
        overlap = int(overlap_percent * window)
    else:
        overlap = window - 1
    return overlap


def sliding_window_view(array: np.ndarray, window_length: int, overlap: int, nan_padding: bool = False) -> np.ndarray:
    """Create a sliding window view of an input array with given window length and overlap.

    .. warning::
       This function will return by default a view onto your input array, modifying values in your result will directly
       affect your input data which might lead to unexpected behaviour! If padding is disabled (default), last window
       fraction of input may not be returned! However, if `nan_padding` is enabled, this will always return a copy
       instead of a view of your input data, independent if padding was actually performed or not!

    Parameters
    ----------
    array : :class:`~numpy.ndarray` with shape (n,) or (n, m)
        array on which sliding window action should be performed. Windowing
        will always be performed along axis 0.
    window_length : int
        length of desired window (must be smaller than array length n)
    overlap : int
        length of desired overlap (must be smaller than window_length)
    nan_padding : bool
        select if last window should be nan-padded or discarded if it not fits with input array length. If nan-padding
        is enabled the return array will always be a copy of the input array independent if padding was actually
        performed or not!

    Returns
    -------
    :class:`~numpy.ndarray`
        windowed view (or copy if ``nan_padding`` is ``True``) of input array as specified,
        last window might be nan-padded if necessary to match window size

    Examples
    --------
    >>> data = np.arange(0,10)
    >>> windowed_view = sliding_window_view(array = data, window_length = 5, overlap = 3, nan_padding = True)
    >>> windowed_view
    array([[ 0.,  1.,  2.,  3.,  4.],
           [ 2.,  3.,  4.,  5.,  6.],
           [ 4.,  5.,  6.,  7.,  8.],
           [ 6.,  7.,  8.,  9., nan]])

    """
    if overlap >= window_length:
        raise ValueError("Invalid Input, overlap must be smaller than window length!")

    if window_length < 2:
        raise ValueError("Invalid Input, window_length must be larger than 1!")

    # calculate length of necessary np.nan-padding to make sure windows and overlaps exactly fits data length
    n_windows = np.ceil((len(array) - window_length) / (window_length - overlap)).astype(int)
    pad_length = window_length + n_windows * (window_length - overlap) - len(array)

    # had to handle 1D arrays separately
    if array.ndim == 1:
        if nan_padding:
            # np.pad always returns a copy of the input array even if pad_length is 0!
            array = np.pad(array.astype(float), (0, pad_length), constant_values=np.nan)

        new_shape = (array.size - window_length + 1, window_length)
    else:
        if nan_padding:
            # np.pad always returns a copy of the input array even if pad_length is 0!
            array = np.pad(array.astype(float), [(0, pad_length), (0, 0)], constant_values=np.nan)

        shape = (window_length, array.shape[-1])
        n = np.array(array.shape)
        o = n - shape + 1  # output shape
        new_shape = np.concatenate((o, shape), axis=0)

    # apply stride_tricks magic
    new_strides = np.concatenate((array.strides, array.strides), axis=0)
    view = np.lib.stride_tricks.as_strided(array, new_shape, new_strides)[0 :: (window_length - overlap)]

    view = np.squeeze(view)  # get rid of single-dimensional entries from the shape of an array.

    return view


def downsample(
    data: arr_t,
    fs_in: float,
    fs_out: float,
) -> np.ndarray:
    """Downsample input signal to a new sampling rate.

    If the output sampling rate is a divisor of the input sampling rate, the signal is downsampled using
    :func:`~scipy.signal.decimate`. Otherwise, data is first filtered using an aliasing filter before it is
    downsampled using linear interpolation.


    Parameters
    ----------
    data : :class:`~numpy.ndarray`
        input data
    fs_in : float
        sampling rate of input data in Hz.
    fs_out : float
        sampling rate of output data in Hz


    Returns
    -------
    :class:`~numpy.ndarray`
        output data with new sampling rate

    """
    data = sanitize_input_nd(data)

    if (fs_in / fs_out) % 1 == 0:
        return signal.decimate(data, int(fs_in / fs_out), axis=0)
    # aliasing filter
    b, a = signal.cheby1(N=8, rp=0.05, Wn=0.8 / (fs_in / fs_out))
    data_lp = signal.filtfilt(a=a, b=b, x=data)
    # interpolation
    x_old = np.linspace(0, len(data_lp), num=len(data_lp), endpoint=False)
    x_new = np.linspace(0, len(data_lp), num=int(len(data_lp) / (fs_in / fs_out)), endpoint=False)
    interpol = interpolate.interp1d(x=x_old, y=data_lp)
    return interpol(x_new)


def _bool_fill(indices: np.ndarray, bool_values: np.ndarray, array: np.ndarray) -> np.ndarray:
    """Fill a pre-allocated array with bool values.

    This function iterates over the indices and adds the bool_values to the array at the given
    indices using a logical "or".

    Parameters
    ----------
    indices : :class:`~numpy.ndarray`
        list with indices to add to the array
    bool_values : :class:`~numpy.ndarray`
        bool values to add to the array
    array : :class:`~numpy.ndarray`
        pre-allocated array to fill with bool values

    Returns
    -------
    :class:`~numpy.ndarray`
        array filled with bool values

    """
    for i in range(len(indices)):  # pylint:disable=consider-using-enumerate
        index = indices[i]
        val = bool_values[i]
        index = index[~np.isnan(index)]
        # perform logical or operation to combine all overlapping window results
        array[index] = np.logical_or(array[index], val)
    return array


def bool_array_to_start_end_array(bool_array: np.ndarray) -> np.ndarray:
    """Find regions in bool array and convert those to start-end indices.

    .. note::
        The end index is inclusive!

    Parameters
    ----------
    bool_array : :class:`~numpy.ndarray` with shape (n,)
        boolean array with either 0/1, 0.0/1.0 or True/False elements

    Returns
    -------
    :class:`~numpy.ndarray`
        array of [start, end] indices with shape (n,2)

    Examples
    --------
    >>> example_array = np.array([0,0,1,1,0,0,1,1,1])
    >>> start_end_list = bool_array_to_start_end_array(example_array)
    >>> start_end_list
    array([[2, 4],
           [6, 9]])
    >>> example_array[start_end_list[0, 0]: start_end_list[0, 1]]
    array([1, 1])

    """
    # check if input is actually a boolean array
    if not np.array_equal(bool_array, bool_array.astype(bool)):
        raise ValueError("Input must be boolean array!")

    slices = np.ma.flatnotmasked_contiguous(np.ma.masked_equal(bool_array, 0))
    return np.array([[s.start, s.stop] for s in slices])


def split_array_equally(data: arr_t, n_splits: int) -> list[tuple[int, int]]:
    """Generate indices to split array into parts with equal lengths.

    Parameters
    ----------
    data : array_like
        data to split
    n_splits : int
        number of splits

    Returns
    -------
    list of tuples
        list with start and end indices which will lead to splitting array into parts with equal lengths

    """
    idx_split = np.arange(0, n_splits + 1) * ((len(data) - 1) // n_splits)
    split_boundaries = list(zip(idx_split[:-1], idx_split[1:]))
    return split_boundaries


def accumulate_array(data: arr_t, fs_in: float, fs_out: float) -> arr_t:
    """Accumulate 1-d array by summing over windows.

    Parameters
    ----------
    data : array_like
        data to accumulate. must be 1-d array
    fs_in : float
        sampling rate of input data in Hz
    fs_out : float
        sampling rate of output data in Hz

    Returns
    -------
    array_like
        accumulated array

    """
    arr = sanitize_input_nd(data, ncols=1)
    n_samples = int(fs_in // fs_out)
    padded_arr = np.pad(arr, (0, n_samples - len(arr) % n_samples), "constant", constant_values=0)
    out = padded_arr.reshape((len(padded_arr) // n_samples, -1)).sum(axis=1)
    if isinstance(data, pd.DataFrame) and isinstance(data.index, pd.DatetimeIndex):
        out = add_datetime_index(out, start_time=data.index[0], sampling_rate=fs_out, column_name=data.columns)
    return out


def add_datetime_index(
    arr: arr_t, start_time: pd.Timestamp, sampling_rate: float, column_name: Optional[str_t] = None
) -> pd.DataFrame:
    """Add datetime index to dataframe.

    Parameters
    ----------
    arr : array_like
        numpy array to add index to
    start_time : :class:`~pandas.Timestamp`
        start time of the index
    sampling_rate : float
        sampling rate of input data in Hz.
    column_name : str, optional
        column of the resulting dataframe or ``None`` to leave it empty.

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with datetime index

    """
    if isinstance(column_name, str):
        # assert list
        column_name = [column_name]
    arr = pd.DataFrame(arr, columns=column_name) if column_name is not None else pd.DataFrame(arr)

    # convert start_time to seconds
    start_time_s = start_time.timestamp()

    arr.index = pd.to_datetime(
        (arr.index * 1 / sampling_rate + start_time_s).astype(int), utc=True, unit="s"
    ).tz_convert(start_time.tzinfo)

    arr.index.name = "time"
    return arr



from typing import Optional

from biopsykit.signals._base_extraction import BaseExtraction

__all__ = ["BaseBPointExtraction"]

from biopsykit.utils.dtypes import BPointDataFrame, CPointDataFrame, HeartbeatSegmentationDataFrame, IcgRawDataFrame


class BaseBPointExtraction(BaseExtraction):
    """Base class for B-point extraction algorithms."""

    points_: BPointDataFrame

    def extract(
        self,
        *,
        icg: IcgRawDataFrame,
        heartbeats: HeartbeatSegmentationDataFrame,
        c_points: CPointDataFrame,
        sampling_rate_hz: Optional[float],
    ):
        raise NotImplementedError("This is an abstract method and needs to be implemented in a subclass.")



"""Module for setting up a pipeline for statistical analysis."""

import re
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, Optional, Union

import pandas as pd
import pingouin as pg

from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_has_index_levels
from biopsykit.utils._types_internal import path_t, str_t

MAP_STAT_TESTS = {
    "normality": pg.normality,
    "equal_var": pg.homoscedasticity,
    "anova": pg.anova,
    "welch_anova": pg.welch_anova,
    "rm_anova": pg.rm_anova,
    "mixed_anova": pg.mixed_anova,
    "ancova": pg.ancova,
    "kruskal": pg.kruskal,
    "friedman": pg.friedman,
    "pairwise_ttests": pg.pairwise_ttests,
    "pairwise_tests": pg.pairwise_tests,
    "pairwise_tukey": pg.pairwise_tukey,
    "pairwise_gameshowell": pg.pairwise_gameshowell,
}

MAP_STAT_PARAMS = {
    "normality": ["dv", "group"],
    "equal_var": ["dv", "group"],
    "anova": ["dv", "between"],
    "welch_anova": ["dv", "between"],
    "rm_anova": ["dv", "within", "subject"],
    "mixed_anova": ["dv", "between", "within", "subject"],
    "ancova": ["dv", "between", "covar"],
    "kruskal": ["dv", "between"],
    "friedman": ["dv", "within", "subject"],
    "pairwise_ttests": ["dv", "between", "within", "subject", "effsize", "tail", "parametric", "padjust"],
    "pairwise_tests": ["dv", "between", "within", "subject", "effsize", "tail", "parametric", "padjust"],
    "pairwise_tukey": ["dv", "between", "effsize"],
    "pairwise_gameshowell": ["dv", "between", "effsize"],
}

MAP_NAMES = {
    "normality": "Test for Normal Distribution",
    "equal_var": "Test for Homoscedasticity (Equal Variances)",
    "anova": "ANOVA",
    "welch_anova": "Welch ANOVA",
    "rm_anova": "Repeated-measurement ANOVA",
    "mixed_anova": "Mixed ANOVA",
    "ancova": "ANCOVA",
    "kruskal": "Kruskal-Wallis H-test for independent samples",
    "friedman": "Friedman test for repeated measurements",
    "pairwise_ttests": "Pairwise t-Tests",
    "pairwise_tests": "Pairwise Tests",
    "pairwise_tukey": "Pairwise Tukey's HSD (Honestly Significant Differences) Test",
    "pairwise_gameshowell": "Pairwise Games-Howell post-hoc Test",
}

MAP_CATEGORIES = {
    "prep": "Preparatory Analysis",
    "test": "Statistical Tests",
    "posthoc": "Post-Hoc Analysis",
}

MAP_LATEX_EXPORT = {
    "T": "t",
    "T_collapse": r"$t({})$",
    "U-val": "U",
    "F_collapse": r"$F({}, {})$",
    "Q_collapse": r"$Q({})$",
    "dof": "df",
    "df1": "$df_{Den}$",
    "df2": "$df_{Nom}$",
    "p-unc": "p",
    "p-corr": "p",
    "p-tukey": "p",
    "pval": "p",
    "np2": r"$\eta^2_p$",
    "np2_row": r"$\eta^2_p$",
    "ng2": r"$\eta^2_G$",
    "ng2_row": r"$\eta^2_G$",
    "n2": r"$\eta^2$",
    "n2_row": r"$\eta^2$",
    "hedges": "Hedges' g",
    "hedges_row": "g",
    "cohen": "Cohen's d",
    "cohen_row": "d",
    "CLES": "CLES",
    "CLES_row": "CLES",
}


STATS_CATEGORY = Literal["prep", "test", "posthoc"]
STATS_EFFECT_TYPE = Literal["between", "within", "interaction"]
PLOT_TYPE = Literal["single", "multi"]

_sig_cols = ["p-corr", "p-tukey", "p-unc", "pval"]
_eff_cols = ["np2", "ng2", "n2", "hedges", "cohen", "CLES"]


class StatsPipeline:
    """Class to set up a pipeline for statistical analysis."""

    def __init__(self, steps: Sequence[tuple[str, str]], params: dict[str, str], **kwargs):
        """Class to set up a pipeline for statistical analysis.

        The purpose of such a pipeline is to assemble several steps of a typical statistical analysis procedure while
        setting different parameters. The parameters passed to this class depend on the used statistical functions.
        It enables setting parameters of the various steps using their names and the parameter name separated by a "__",
        as in the examples below.

        The interface of this class is inspired by the scikit-learn Pipeline for machine learning tasks
        (:class:`~sklearn.pipeline.Pipeline`).

        All functions used are from the ``pingouin`` library (https://pingouin-stats.org/) for statistical analysis.

        The different steps of statistical analysis can be divided into categories:

        * *Preparatory Analysis* (``prep``): Analyses applied to the data before performing the actual statistical
          analysis. Currently supported functions are:

          * ``normality``: Test whether a random sample comes from a normal distribution.
            See :func:`~pingouin.normality` for further information.
          * ``equal_var``: Test equality of variances (homoscedasticity).
            See :func:`~pingouin.homoscedasticity` for further information.

        * *Statistical Test* (``test``): Statistical test to determine differences or similarities in the data.
          Currently supported functions are:

            * ``pairwise_tests``: Pairwise tests (either for independent or dependent samples).
              See :func:`~pingouin.pairwise_tests` for further information.
            * ``anova``: One-way or N-way ANOVA. See :func:`~pingouin.anova` for further information.
            * ``welch_anova``: One-way Welch-ANOVA. See :func:`~pingouin.welch_anova` for further
              information.
            * ``rm_anova``: One-way and two-way repeated measures ANOVA. See :func:`~pingouin.rm_anova`
              further information.
            * ``mixed_anova``: Mixed-design (split-plot) ANOVA. See :func:`~pingouin.mixed_anova` for
              further information.
            * ``kruskal``: Kruskal-Wallis H-test for independent samples. See :func:`~pingouin.kruskal`
              for further information.

        * *Posthoc Tests* (``posthoc``): Posthoc tests to determine differences of individual groups if more than two
          groups are analyzed.
          Currently supported functions are:

          * ``pairwise_tests``: Pairwise tests (either for independent or dependent samples).
            See :func:`~pingouin.pairwise_tests` for further information.
          * ``pairwise_tukey``: Pairwise Tukey-HSD post-hoc test.
            See :func:`~pingouin.pairwise_tukey` for further information.
          * ``pairwise_gameshowell``: Pairwise Games-Howell post-hoc test.
            See :func:`~pingouin.pairwise_gameshowell` for further information.

        A ``StatsPipeline`` consists of a list of tuples specifying the individual ``steps`` of the pipeline.
        The first value of each tuple indicates the category this step belongs to (``prep``, ``test``, or ``posthoc``),
        the second value indicates the analysis function to use in this step (e.g., ``normality``, or ``rm_anova``).

        Furthermore, a ``params`` dictionary specifying the parameters and variables for statistical analysis
        needs to be supplied. Parameters can either be specified *globally*, i.e., for all steps in the pipeline
        (the default), or *locally*, i.e., only for one specific category, by prepending the category and separating it
        from the parameter name by a `__`. The parameters depend on the type of analysis used in the pipeline.

        Examples are:

        * ``dv``: column name of the dependent variable
        * ``between``: column name of the between-subject factor
        * ``within``: column name of the within-subject factor
        * ``effsize``: type of effect size to compute (if applicable)
        * ``multicomp``: whether (and how) to apply multi-comparison correction of p-values to the *last* step in the
          pipeline (either "test" or "posthoc") using :meth:`~biopsykit.stats.StatsPipeline.multicomp`.
          The arguments for the call to :meth:`~biopsykit.stats.StatsPipeline.multicomp` are supplied via dictionary.
        * ...

        Parameters
        ----------
        steps : list of tuples
            list of tuples specifying statistical analysis pipeline
        params : dict
            dictionary with parameter names and their values
        **kwargs : dict
            additional arguments, such as:

            * ``round``: Set the default decimal rounding of the output dataframes or ``None`` to disable rounding.
              Default: Rounding to 4 digits on p-value columns only. See :meth:`~pandas.DataFrame.round` for further
              options.

        """
        self.steps = steps
        self.params = params
        self.data: Optional[pd.DataFrame] = None
        self.results: dict[str, pd.DataFrame] = {}
        self.category_steps = {}
        self.round = kwargs.get("round", {col: 4 for col in _sig_cols})
        for step in self.steps:
            self.category_steps.setdefault(step[0], [])
            self.category_steps[step[0]].append(step[1])

    def apply(self, data: pd.DataFrame) -> dict[str, pd.DataFrame]:
        """Apply statistical analysis pipeline on input data.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data to apply statistical analysis pipeline on. Must be provided in *long-format*.

        Returns
        -------
        dict
            dictionary with results from all pipeline steps

        """
        self.data = data
        pipeline_results = {}
        data = data.reset_index()

        for i, step in enumerate(self.steps):
            general_params = {key: value for key, value in self.params.items() if len(key.split("__")) == 1}
            specific_params = {
                key.split("__")[1]: value
                for key, value in self.params.items()
                if len(key.split("__")) > 1 and step[0] in key.split("__")
            }
            params = {key: general_params[key] for key in MAP_STAT_PARAMS[step[1]] if key in general_params}

            grouper = []
            grouper_tmp = self._get_grouper_variable(general_params, specific_params)
            grouper = grouper + grouper_tmp

            if step[0] == "prep":
                grouper, specific_params = self._get_specific_params_prep(grouper, general_params, specific_params)

            test_func = MAP_STAT_TESTS[step[1]]
            if len(grouper) > 0:
                result = data.groupby(grouper, sort=general_params.get("sort", True)).apply(
                    lambda df, tf=test_func, sp=specific_params, p=params: tf(data=df, **sp, **p)
                )
            else:
                result = test_func(data=data, **specific_params, **params)

            if i == len(self.steps) - 1 and "multicomp" in general_params:
                # apply multi-comparison correction (p-adjustment) for the last analysis step in the pipeline
                # if it was enabled
                multicomp_dict = general_params.get("multicomp")
                if multicomp_dict is None:
                    multicomp_dict = {}
                result = self.multicomp(result, **multicomp_dict)

            pipeline_results[step[1]] = result

        self.results = pipeline_results
        return pipeline_results

    @staticmethod
    def _get_grouper_variable(general_params: dict[str, str], specific_params: dict[str, str]):
        grouper_tmp = []
        if "groupby" in specific_params:
            grouper_tmp = specific_params.pop("groupby")
        elif "groupby" in general_params:
            grouper_tmp = general_params.pop("groupby")

        if isinstance(grouper_tmp, str):
            grouper_tmp = [grouper_tmp]

        return grouper_tmp

    @staticmethod
    def _get_specific_params_prep(grouper: list[str], general_params: dict[str, str], specific_params: dict[str, str]):
        if "within" in general_params and "between" in general_params:
            grouper.append(general_params["within"])
            specific_params["group"] = general_params["between"]
        else:
            specific_params["group"] = general_params.get("within", general_params.get("between"))

        return grouper, specific_params

    def results_cat(self, category: STATS_CATEGORY) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
        """Return results for pipeline category.

        This function filters results from steps belonging to the specified category and returns them.

        Parameters
        ----------
        category : {"prep", "test", "posthoc"}
            category name

        Returns
        -------
        :class:`~pandas.DataFrame` or dict
            dataframe with results from the specified category or dict of such if multiple steps belong to the
            same category.

        """
        cat = self.category_steps.get(category, [])
        if len(cat) == 1:
            return self.results[cat[0]]
        if len(cat) > 1:
            return {c: self.results[c] for c in cat}
        return {}

    def _ipython_display_(self):
        try:
            from IPython.core.display import display  # pylint:disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                "Displaying statistics results failed because "
                "IPython cannot be imported. Install it via 'pip install ipython'."
            ) from e

        display(self._param_df().T)
        display(self._result_df().T)

    def display_results(self, sig_only: Optional[Union[str, bool, Sequence[str], dict[str, bool]]] = None, **kwargs):
        """Display formatted results of statistical analysis pipeline.

        This function displays the results of the statistical analysis pipeline. The output is Markdown-formatted and
        optimized for Jupyter Notebooks. The output can be configured to for example:

        * only show specific categories
        * only show statistically significant results for specific categories or for all pipeline categories
        * display results grouped by a grouper (when pipeline is applied on multiple groups of data, e.g.,
          on multiple feature of the same type independently)

        Parameters
        ----------
        sig_only : bool, str, list, or dict, optional
            whether to only show statistically significant (p < 0.05) results or not. ``sig_only`` accepts multiple
            possible formats:

            * ``str``: filter only one specific category or "all" to filter all categories by statistical
              significance
            * ``bool``: ``True`` to filter all categories by statistical significance, ``False`` otherwise
            * ``list``: list of categories whose results should be filtered by statistical significance
            * ``dict``: dictionary with category names and bool values to filter (or not) for statistical
              significance

            Default: ``None`` (no filtering)
        **kwargs
            additional arguments to be passed to the function, such as:

            * ``category`` names: ``True`` to display results of this category, ``False`` to skip displaying results
              of this category. Default: show results from all categories
            * ``grouped``: ``True`` to group results by the variable "groupby" specified in the parameter
              dictionary when initializing the ``StatsPipeline`` instance.

        """
        try:
            from IPython.core.display import Markdown, display  # pylint:disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                "Displaying statistics results failed because "
                "IPython cannot be imported. Install it via 'pip install ipython'."
            ) from e

        sig_only = self._get_sig_only(sig_only)
        grouped = kwargs.pop("grouped", False)

        if self.results is None:
            display(Markdown("No results."))
            return

        if grouped and "groupby" in self.params:
            for key, _ in self.data.groupby(self.params.get("groupby")):
                display(Markdown(f"""<font size="4"><b> {key} </b></font>"""))
                self._display_results(sig_only, self.params.get("groupby"), key, **kwargs)
        else:
            self._display_results(sig_only, **kwargs)

    def _get_sig_only(self, sig_only: Optional[Union[str, bool, Sequence[str], dict[str, bool]]] = None):
        if sig_only is None:
            sig_only = {}
        if isinstance(sig_only, str):
            sig_only = True if sig_only == "all" else [sig_only]
        if isinstance(sig_only, bool):
            sig_only = {cat: sig_only for cat in self.category_steps}
        if isinstance(sig_only, list):
            sig_only = {cat: cat in sig_only for cat in self.category_steps}
        return sig_only

    def _display_results(
        self, sig_only: dict[str, bool], groupby: Optional[str] = None, group_key: Optional[str] = None, **kwargs
    ):
        try:
            from IPython.core.display import Markdown, display  # pylint:disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                "Displaying statistics results failed because "
                "IPython cannot be imported. Install it via 'pip install ipython'."
            ) from e

        display(Markdown("""<font size="3"><b> Overview </b></font>"""))
        display(self)
        for category, steps in self.category_steps.items():
            if kwargs.get(category, True):
                self._display_category(category, steps, sig_only, groupby, group_key)

    @staticmethod
    def _filter_sig(data: pd.DataFrame) -> Optional[pd.DataFrame]:
        for col in _sig_cols:
            if col in data.columns:
                if data[col].isna().all():
                    # drop column if all values are NaN => most probably because we turned on p-adjust but only
                    # have two main effects
                    data = data.drop(columns=col)
                    continue
                return data[data[col] < 0.05]
        return None

    @staticmethod
    def _filter_pcol(data: Union[pd.DataFrame, pd.Series]) -> Optional[pd.Series]:
        for col in _sig_cols:
            if isinstance(data, pd.DataFrame) and col in data.columns:
                return data[col]
            if isinstance(data, pd.Series) and col in data.index:
                return data[[col]]
        return None

    @staticmethod
    def _filter_eff_col(data: Union[pd.DataFrame, pd.Series]) -> Optional[pd.Series]:
        for col in _eff_cols:
            if isinstance(data, pd.DataFrame) and col in data.columns:
                return data[col]
            if isinstance(data, pd.Series) and col in data.index:
                return data[[col]]
        return None

    @staticmethod
    def _get_effsize_name(data: Union[pd.DataFrame, pd.Series]) -> Optional[str]:
        for col in _eff_cols:
            if isinstance(data, pd.DataFrame) and col in data.columns:
                return col
            if isinstance(data, pd.Series) and col in data.index:
                return col
        return None

    def _filter_effect(self, stats_category: STATS_CATEGORY, stats_effect_type: STATS_EFFECT_TYPE) -> pd.DataFrame:
        results = self.results_cat(stats_category)
        if len(results) == 0:
            raise ValueError(f"No results for category {stats_category}!")
        if "Contrast" in results.columns:
            if stats_effect_type == "interaction":
                if "between" in self.params and "within" in self.params:
                    # interaction
                    effect1 = self.params["within"]
                    effect2 = self.params["between"]
                elif "between" in self.params:
                    # two-level between
                    effect1 = self.params["between"][0]
                    effect2 = self.params["between"][1]
                else:
                    # two-level within
                    effect1 = self.params["within"][0]
                    effect2 = self.params["within"][1]
                key = f"{effect1} * {effect2}"
            else:
                key = self.params[stats_effect_type]

            results = results[results["Contrast"] == key]
            results = results.drop(columns="Contrast")

        return results

    def export_statistics(self, file_path: path_t):
        """Export results of statistics analysis pipeline to Excel file.

        Each step of the analysis pipeline is saved into its own sheet. The first sheet is an overview of the
        parameters specified in the analysis pipeline.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            path to export file

        """
        # assert pathlib
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".xlsx")

        writer = pd.ExcelWriter(file_path, engine="xlsxwriter")  # pylint:disable=abstract-class-instantiated
        workbook = writer.book
        header_format = workbook.add_format({"bold": True})
        param_df = self._param_df()
        param_df.to_excel(writer, sheet_name="parameter")
        for key, df in self.results.items():
            df.to_excel(writer, sheet_name=key, startrow=1)
            worksheet = writer.sheets[key]
            worksheet.write_string(0, 0, MAP_NAMES[key], header_format)
        writer.close()

    def sig_brackets(
        self,
        stats_category_or_data: Union[STATS_CATEGORY, pd.DataFrame],
        stats_effect_type: STATS_EFFECT_TYPE,
        stats_type: Optional[STATS_EFFECT_TYPE] = None,
        plot_type: Optional[PLOT_TYPE] = "single",
        features: Optional[Union[str, Sequence[str], dict[str, Union[str, Sequence[str]]]]] = None,
        x: Optional[str] = None,
        subplots: Optional[bool] = False,
    ) -> Union[
        tuple[Sequence[tuple[str, str]], Sequence[float]],
        tuple[dict[str, Sequence[tuple[tuple[str, str], tuple[str, str]]]], dict[str, Sequence[float]]],
    ]:
        """Generate *significance brackets* used indicate statistical significance in boxplots.

        Parameters
        ----------
        stats_category_or_data : {"prep", "test", "posthoc"} or :class:`~pandas.DataFrame`
            either a string to specify the pipeline category to use for generating significance brackets or a
            dataframe with statistical results if significance brackets should be generated from the dataframe.
        stats_effect_type : {"between", "within", "interaction"}
            type of statistical effect ("between", "within", or "interaction"). Needed to extract the correct
            information from the analysis dataframe.
        stats_type : {"between", "within", "interaction"}
            .. note:: Deprecated in 0.4.0
                `stats_type` will be removed in 0.5.0, it is replaced by `stats_effect_type`.
        plot_type : {"single", "multi"}
            type of plot for which significance brackets are generated: "multi" if boxplots are grouped
            (by ``hue`` variable), "single" (the default) otherwise.
        features : str, list or dict, optional
            feature(s) used in boxplot. The resulting significance brackets will be filtered accordingly to only
            contain features present in the boxplot. It can have the following formats:

            * ``str``: only one feature is plotted in the boxplot
              (returns significance brackets of only one feature)
            * ``list``: multiple features are combined into *one* :class:`~matplotlib.axes.Axes` object
              (returns significance brackets of multiple features)
            * ``dict``: dictionary with feature (or list of features) per subplot if boxplots are structured in
              subplots (``subplots`` is ``True``) (returns dictionary with significance brackets per subplot)

            Default: ``None`` to return significance brackets of all features
        x : str, optional
            name of ``x`` variable used to plot data in boxplot. Only required if ``plot_type`` is "multi".
        subplots : bool, optional
            ``True`` if multiple boxplots are structured in subplots, ``False`` otherwise. Default: ``False``

        Returns
        -------
        box_pairs
            list with significance brackets (or dict of such if ``subplots`` is ``True``)
        pvalues
            list with p values belonging to the significance brackets in ``box_pairs`` (or dict of such if
            ``subplots`` is ``True``)

        """
        if stats_type is not None:
            warnings.warn(
                "Argument 'stats_type' is deprecated in 0.4.0 and was replaced by 'stats_effect_type'. "
                "It will be removed in 0.5.0.",
                category=DeprecationWarning,
            )
            stats_effect_type = stats_type

        stats_data = self._extract_stats_data(stats_category_or_data, stats_effect_type)

        # if features is None:
        #     features = list(stats_data.index)
        # features = self._sanitize_features_input(features)

        if stats_effect_type == "interaction":
            stats_data, box_pairs = self._get_stats_data_box_pairs_interaction(stats_data)
        else:
            stats_data, box_pairs = self._get_stats_data_box_pairs(stats_data, plot_type, features, x)

        if box_pairs.empty:
            return [], []

        pvalues = self._filter_pcol(stats_data)
        if subplots:
            return self._sig_brackets_dict(box_pairs, pvalues, features)

        return list(box_pairs), list(pvalues)

    @staticmethod
    def _sig_brackets_dict(
        box_pairs: pd.Series,
        pvalues: pd.Series,
        features: Union[Sequence, dict[str, Union[str, Sequence[str]]]],
    ) -> tuple[dict[str, Sequence[tuple[tuple[str, str], tuple[str, str]]]], dict[str, Sequence[float]]]:
        dict_box_pairs = {}
        dict_pvalues = {}
        if features is None:
            features = list(box_pairs.index.unique())
        if isinstance(features, list):
            features = {f: f for f in features}
        for key in features:
            features_list = features[key]

            if isinstance(features_list, str):
                features_list = [features_list]
            list_pairs = dict_box_pairs.setdefault(key, [])
            list_pvalues = dict_pvalues.setdefault(key, [])
            for i, (idx, sig_pair) in enumerate(box_pairs.items()):
                if str(idx) in str(features_list):
                    list_pairs.append(sig_pair)
                    list_pvalues.append(pvalues[i])

        return dict_box_pairs, dict_pvalues

    def _param_df(self):
        return pd.DataFrame(
            [str(s) for s in self.params.values()],
            index=self.params.keys(),
            columns=["parameter"],
        )

    def _result_df(self):
        return pd.DataFrame(
            [s[1] for s in self.steps],
            index=[s[0] for s in self.steps],
            columns=["parameter"],
        )

    def stats_to_latex(
        self,
        stats_test: Optional[str] = None,
        index: Optional[Union[tuple, str]] = None,
        data: Optional[pd.Series] = None,
    ) -> str:
        """Generate LaTeX output from statistical results.

        Parameters
        ----------
        stats_test : str, optional
            name of statistical test in ``StatsPipeline``, e.g., "pairwise_tests" or "anova" or ``None``
            if external statistical results is provided via ``data``
        index : str or tuple, optional
            row indexer of statistical result or ``None`` to generate LaTeX output for all rows.
            Default: ``None``
        data : :class:`~pandas.DataFrame`, optional
            dataframe with optional external statistical results

        Returns
        -------
        str
            LaTeX output that can be copied and pasted into LaTeX documents

        Raises
        ------
        ValueError
            if both ``data`` and ``stats_test`` is ``None``

        """
        if all(param is None for param in [stats_test, data]):
            raise ValueError("Either 'data' or 'stats_test' must be provided!")
        if stats_test is not None:
            data = self.results[stats_test].copy()
        if index is not None:
            data = data.loc[index].squeeze()

        if isinstance(data, pd.DataFrame):
            data = self._stats_to_index_set_index(data)
            return data.apply(self._stats_to_latex_row, axis=1)
        return self._stats_to_latex_row(data)

    def _stats_to_latex_row(self, row: pd.Series) -> str:
        pval = self._format_pval(row)
        eff_name = self._get_effsize_name(row)
        eff_name_print = f"{eff_name}_row"
        if "T" in row:
            dof = self._format_dof(row["dof"])
            tval = self._format_number(row["T"])
            effsize = self._format_number(row[eff_name])
            return f"$t({dof}) = {tval}, p {pval}, {MAP_LATEX_EXPORT[eff_name_print]} = {effsize}$"
        if "F" in row:
            rename_dict = {"ddof1": "df1", "ddof2": "df2", "DF": "df", "DF1": "df1", "DF2": "df2"}
            row = row.rename(rename_dict)
            dofs = (row["df1"], row["df2"]) if "df1" in row else (row["df"],)
            dofs = [self._format_dof(dof) for dof in dofs]
            dofs = ",".join(dofs)
            fval = self._format_number(row["F"])
            ret_string = rf"$F({dofs}) = {fval}, p {pval}$"
            if eff_name is not None:
                effsize = self._format_number(row[eff_name])
                ret_string = ret_string[:-1] + rf", {MAP_LATEX_EXPORT[eff_name_print][1:-1]} = {effsize}$"
            return ret_string
        if "U-val" in row:
            effsize = self._format_number(row[eff_name])
            uval = self._format_number(row["U-val"])
            return f"$U = {uval}, p {pval}, {MAP_LATEX_EXPORT[eff_name_print]} = {effsize}$"
        if "W-val" in row:
            effsize = self._format_number(row[eff_name])
            wval = self._format_number(row["W-val"])
            return f"$W = {wval}, p {pval}, {MAP_LATEX_EXPORT[eff_name_print]} = {effsize}$"
        return ""

    def _format_pval(self, row: pd.Series) -> str:
        pval = round(self._filter_pcol(row)[0], 3)
        if pval < 0.001:
            pval = "< 0.001"
        elif pval > 0.999:
            pval = "> 0.999"
        else:
            pval = f"= {pval:.3f}"
        return pval

    def results_to_latex_table(  # pylint:disable=too-many-branches
        self,
        stats_test: str,
        data: Optional[pd.DataFrame] = None,
        stats_effect_type: Optional[STATS_EFFECT_TYPE] = None,
        unstack_levels: Optional[str_t] = None,
        collapse_dof: Optional[bool] = True,
        si_table_format: Optional[str] = None,
        index_kws: Optional[dict[str, Any]] = None,
        column_kws: Optional[dict[str, Any]] = None,
        show_a_b: Optional[bool] = False,
        **kwargs,
    ) -> str:
        r"""Convert statistical result dataframe to LaTeX table.

        This function converts a dataframe from a statistical analysis to a LaTeX table using
        :meth:`~pandas.DataFrame.to_latex`.

        This function uses the LaTeX package ``siunitx`` (https://ctan.org/pkg/siunitx?lang=en) to represent numbers.
        By default, the column format for columns that contain numbers is "S" which is provided by ``siunitx``.
        The column format can be configured by the ``si_table_format`` argument.


        Parameters
        ----------
        stats_test : str, optional
            name of statistical test in ``StatsPipeline``, e.g., "pairwise_tests" or "anova".
        data : :class:`~pandas.DataFrame`, optional
            dataframe with optional external statistical results
        stats_effect_type : {"between", "within", "interaction"}
            type of statistical effect ("between", "within", or "interaction"). Needed to extract the correct
            information from the analysis dataframe.
        unstack_levels : str or list of str, optional
            name(s) of dataframe index level(s) to be unstacked in the resulting latex table or ``None``
            to unstack no level(s)
        collapse_dof : bool, optional
            ``True`` to collapse degree-of-freedom (dof) from a separate column into the column header of the
            t- or F-value, respectively, ``False`` to keep it as separate "dof" column. This only works if
            the degrees-of-freedom are the same for all tests in the table. Default: ``True``
        si_table_format : str, optional
            table format for the numbers in the LaTeX table.
        index_kws : dict, optional
            dictionary containing arguments to configure how the table index is formatted. Possible arguments are:

            * index_italic : bool
              ``True`` to format index columns in italic, ``False`` otherwise. Default: ``True``
            * index_level_order : list
              list of index level names indicating the index level order of a :class:`~pandas.MultiIndex`
              in the LaTeX table. If `None` the index order of the dataframe will be used
            * index_value_order :  list or dict
              list of index values if rows in LaTeX table should have a different order than the underlying
              dataframe or if only specific rows should be exported as LaTeX table. If the table index is a
              :class:`~pandas.MultiIndex` then ``index_value_order`` should be a dictionary with the index level
              names as keys and lists of index values of the specific level as values
            * index_rename_map : dict
                mapping with dictionary with index values as keys and new index values to be exported
            * index_level_names_tex : str of list of str
                names of index levels in the LaTeX table or ``None`` to keep the index level names of the dataframe
        column_kws : dict, optional
            dictionary containing arguments to configure how the table columns are formatted. Possible arguments are:

            * column_level_order : list
              list of column level names indicating the column level order of a :class:`~pandas.MultiIndex`
              in the LaTeX table. If `None` the order of the dataframe will be used
            * column_value_order :  list or dict
              list of column values if columns in LaTeX table should have a different order than the underlying
              dataframe or if only specific columns should be exported as LaTeX table. If the table column index is a
              :class:`~pandas.MultiIndex` then ``column_value_order`` should be a dictionary with the column level
              names as keys and lists of column values of the specific level as values
            * column_rename_map : dict
                mapping with dictionary with column values as keys and new column values to be exported
            * column_level_names_tex : str of list of str
                names of column levels in the LaTeX table or ``None`` to keep the column level names of the dataframe
        show_a_b : bool, optional
            ``True`` to add the names of the measurements (columns "A" and "B") to the output table,
            ``False`` otherwise. Only needed for pairwise tests (:func:`~pingouin.pairwise_tests`). Default: ``False``
        kwargs
            additional keywords that are passed to :meth:`~pandas.DataFrame.to_latex`.
            The following default arguments will be passed if not specified otherwise:

            * column_format: str
              The columns format as specified in LaTeX table format e.g. "rcl" for 3 columns. By default, the column
              format is automatically inferred from the dataframe, with index columns being formatted as "l" and
              value columns formatted as "S". If column headers are multi-columns, "|" will be added as separators
              between each group.
            * multicolumn_format : str
              The alignment for multi-columns. Default: "c"
            * escape : bool
              By default, the value will be read from the pandas config module. When set to ``False`` prevents from
              escaping latex special characters in column names. Default: ``False``
            * position : str
              The LaTeX positional argument for tables, to be placed after ``\begin{}`` in the output. Default: "th!"


        Returns
        -------
        str
            LaTeX code of formatted table

        """
        if data is None:
            data = self.results[stats_test].copy()

        if stats_effect_type is not None:
            data = data.set_index("Source", append=True)
            try:
                data = data.xs(stats_effect_type, level="Source")
            except KeyError:
                data = data.xs(self.params[stats_effect_type], level="Source")

        if si_table_format is None:
            si_table_format = "table-format = <1.3"

        if index_kws is None:
            index_kws = {}
        if column_kws is None:
            column_kws = {}

        kwargs.setdefault("multicolumn_format", "c")
        kwargs.setdefault("escape", False)
        kwargs.setdefault("position", "th!")

        pcol = str(self._filter_pcol(data).name)
        data = self._extract_data_latex_table(
            data, stats_test=stats_test, pcol=pcol, collapse_dof=collapse_dof, show_a_b=show_a_b
        )

        column_map = {col: MAP_LATEX_EXPORT[col] for col in data.columns if col in MAP_LATEX_EXPORT}

        if len(data.index.names) > 1:
            data.index = data.index.droplevel(-1)
        data = data.assign(**{pcol: data.loc[:, pcol].apply(self._format_pvals_stars)})
        data = data.map(self._format_number)

        if unstack_levels is not None:
            data = data.stack()
            data = data.unstack(unstack_levels)
            data = data.unstack(-1)
        data = data.rename(columns=column_map)

        data = self._format_latex_table_column(data, column_kws)
        data = self._format_latex_table_index(data, index_kws, show_a_b)

        kwargs.setdefault("column_format", self._format_latex_column_format(data))

        data_latex = data.to_latex(**kwargs)
        return self._apply_latex_code_correction(data_latex, si_table_format)

    @staticmethod
    def _format_number(val) -> str:
        if isinstance(val, float):
            if val % 1 == 0:
                return str(int(val))
            return f"{val:.3f}"
        return val

    def multicomp(
        self,
        stats_category_or_data: Union[STATS_CATEGORY, pd.DataFrame],
        levels: Optional[Union[bool, str, Sequence[str]]] = False,
        method: Optional[str] = "bonf",
    ) -> pd.DataFrame:
        """Apply multi-comparison correction to results from statistical analysis.

        This function will add a new column ``p-corr`` to the dataframe which contains the adjusted p-values.
        The level(s) on which to perform multi-comparison correction on can be specified by the ``levels`` parameter.

        Parameters
        ----------
        stats_category_or_data : :class:`~pandas.DataFrame`
            dataframe with results from statistical analysis
        levels: bool, str, or list of str, optional
            index level(s) on which to perform multi-comparison correction on, ``True`` to perform multi-comparison
            correction on the *whole* dataset (i.e., on *no* particular index level), or ``False`` or ``None`` to
            perform multi-comparison correction on **all** index levels.
            Default: ``False``
        method : str, optional
            method used for testing and adjustment of p-values. See :func:`~pingouin.multicomp` for the
            available methods. Default: "bonf"

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with adjusted p-values

        """
        if isinstance(stats_category_or_data, pd.DataFrame):
            data = stats_category_or_data
        else:
            data = self.results_cat(stats_category_or_data)

        levels = self._multicomp_get_levels(levels, data)

        _assert_has_index_levels(data, levels, match_atleast=True)

        group_cols = list(data.index.names)[:-1]
        group_cols = list(set(group_cols) - set(levels))

        if len(group_cols) == 0:
            return self._multicomp_lambda(data, method=method)
        return data.groupby(group_cols, group_keys=False).apply(lambda df: self._multicomp_lambda(df, method=method))

    @classmethod
    def _multicomp_get_levels(cls, levels: Union[bool, str, Sequence[str]], data: pd.DataFrame) -> Sequence[str]:
        if levels is None:
            levels = []
        elif isinstance(levels, bool):
            levels = list(data.index.names)[:-1] if levels else []
        elif isinstance(levels, str):
            levels = [levels]
        return levels

    @staticmethod
    def _multicomp_lambda(data: pd.DataFrame, method: str) -> pd.DataFrame:
        for col in list(reversed(_sig_cols[1:])):
            # iterate possible sig_cols in reserved order, except for 'p-corr'
            if col in data.columns:
                data["p-corr"] = pg.multicomp(list(data[col]), method=method)[1]
                break
        return data

    def _extract_stats_data(
        self, stats_category_or_data: Union[STATS_CATEGORY, pd.DataFrame], stats_effect_type: STATS_EFFECT_TYPE
    ):
        if isinstance(stats_category_or_data, (str, pd.DataFrame)):
            if isinstance(stats_category_or_data, str):
                stats_data = self._filter_effect(stats_category_or_data, stats_effect_type)
            else:
                stats_data = stats_category_or_data
        else:
            raise TypeError(
                "Either string with stats category (e.g., 'test' or 'posthoc') or dataframe with stats results must "
                "be supplied as parameter! "
            )
        return self._filter_sig(stats_data)

    @staticmethod
    def _sanitize_features_input(features: Union[str, Sequence[str], dict[str, Union[str, Sequence[str]]]]):
        if isinstance(features, str):
            features = [features]
        if isinstance(features, dict):
            for key, value in features.items():
                # ensure that all entries in the dict are lists for consistency
                if isinstance(value, str):
                    features[key] = [value]

        return features

    def _get_stats_data_box_pairs(
        self,
        stats_data: pd.DataFrame,
        plot_type: Optional[PLOT_TYPE] = "single",
        features: Optional[Union[Sequence[str], dict[str, Union[str, Sequence[str]]]]] = None,
        x: Optional[str] = None,
    ):
        if features is not None:
            if isinstance(features, dict):
                # flatten dict values into list of str
                features = list(features.values())
                features = [item for sublist in features for item in sublist]
            if all(isinstance(feature, str) for feature in features):
                if isinstance(features, str):
                    features = [features]
                stats_data = pd.concat(stats_data.filter(like=f, axis=0) for f in features)
            else:
                stats_data = stats_data.unstack().loc[features].stack()

        stats_data = stats_data.drop_duplicates()
        stats_data = stats_data.reset_index()
        if plot_type == "single":
            box_pairs = self._get_box_pairs_single(stats_data)
        else:
            box_pairs = self._get_box_pairs_multi(stats_data, x)

        return stats_data, box_pairs

    def _get_stats_data_box_pairs_interaction(self, stats_data: pd.DataFrame):
        stats_data = stats_data.reset_index()

        within_param = self.params["within"][0] if not isinstance(self.params["within"], str) else self.params["within"]

        stats_data = stats_data[stats_data[within_param] != "-"]
        index = stats_data[self.params.get("groupby", [])]
        stats_data = stats_data.set_index(within_param)

        box_pairs = stats_data.apply(lambda row: ((row.name, row["A"]), (row.name, row["B"])), axis=1)
        if not index.empty:
            box_pairs.index = index
        return stats_data, box_pairs

    def _get_box_pairs_single(self, stats_data: pd.DataFrame):
        try:
            box_pairs = stats_data.apply(lambda row: (row["A"], row["B"]), axis=1)
        except KeyError as e:
            raise ValueError(
                "Generating significance brackets failed. If ANOVA (or such) was used as "
                "statistical test, significance brackets need to be generated from post-hoc tests!"
            ) from e
        index = stats_data[self.params.get("groupby", [])]
        if not index.empty:
            box_pairs.index = index
        return box_pairs

    def _get_box_pairs_multi(self, stats_data: pd.DataFrame, x: str):
        if x is None:
            raise ValueError("'x' must be specified when 'plot_type' is 'multi'!")
        if x in stats_data.columns:
            stats_data = stats_data.set_index(x)
        else:
            stats_data = stats_data.set_index(self.params["groupby"])
        return stats_data.apply(lambda row: ((row.name, row["A"]), (row.name, row["B"])), axis=1)

    def _display_category(  # pylint:disable=too-many-branches
        self, category: str, steps: Sequence[str], sig_only: dict[str, bool], groupby: str, group_key: str
    ):
        try:
            from IPython.core.display import Markdown, display  # pylint:disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                "Displaying statistics results failed because "
                "IPython cannot be imported. Install it via 'pip install ipython'."
            ) from e
        display(Markdown(f"""<font size="3"><b> {MAP_CATEGORIES[category]} </b></font>"""))
        for step in steps:
            display(Markdown(f"**{MAP_NAMES[step]}**"))
            df = self.results[step]
            if groupby is not None:
                df = df.xs(group_key, level=groupby)
            if sig_only.get(category, False):
                df = self._filter_sig(df)
                if df.empty:
                    display(Markdown("*No significant p-values.*"))
                    continue
            if self.round is None:
                display(df)
            else:
                display(df.round(self.round))

    @staticmethod
    def _format_pvals_stars(pval: float) -> str:
        pstar = pd.cut([pval], [0.0, 0.001, 0.01, 0.05, 1.1], right=False, labels=["***", "**", "*", ""])
        pstar = pstar[0]
        ret = f"{pval:.3f}"
        if pval == 1.0:
            ret = ">0.999"
        if len(pstar) >= 0:
            if pval < 0.001:
                ret = "<0.001"
            ret += rf"$^{{{pstar}}}$"
        return ret

    @staticmethod
    def _format_dof(dof: int) -> str:
        return str(int(dof)) if dof % 1 == 0 else f"{dof:.2f}"

    def _extract_data_ttest(self, data: pd.DataFrame, pcol: str, collapse_dof: bool, show_a_b: bool) -> pd.DataFrame:
        effsize_name = self._get_effsize_name(data)
        columns = ["T", "dof", pcol, effsize_name]
        if show_a_b and "A" in data.columns:
            columns += ["A", "B"]
        data = data[columns]
        if collapse_dof:
            dof = data["dof"].unique()
            if len(dof) != 1:
                raise ValueError(f"Cannot collapse dof in table: dof are not unique! Got {dof}")
            data = data.rename(columns={"T": MAP_LATEX_EXPORT["T_collapse"].format(self._format_dof(dof[0]))})
            data = data.drop(columns="dof")
        if show_a_b and "A" in data.columns:
            data = self._remove_a_b(data)
        return data

    def _extract_data_anova(self, data: pd.DataFrame, pcol: str, collapse_dof: bool) -> pd.DataFrame:
        rename_dict = {"ddof1": "df1", "ddof2": "df2", "DF": "df", "DF1": "df1", "DF2": "df2"}
        data = data.rename(columns=rename_dict)
        columns = []
        if collapse_dof:
            dof_cols = ["df1", "df2"] if "df1" in data.columns else ["df"]
            dofs = tuple(data[col].unique() for col in dof_cols)
            if any(len(d) != 1 for d in dofs):
                raise ValueError(f"Cannot collapse dof in table: dof are not unique! Got {dofs}.")
            dofs = [self._format_dof(dof[0]) for dof in dofs]
            f_col = MAP_LATEX_EXPORT["F_collapse"].format(*dofs)
            columns.append(f_col)
            data = data.rename(columns={"F": f_col})
            data = data.drop(columns=["df1", "df2", "df"], errors="ignore")
        else:
            if "df1" in data.columns:
                data["df"] = "{" + data["df1"].astype(str) + ", " + data["df2"].astype(str) + "}"
                data = data.drop(columns=["df1", "df2"])
            columns.append("df")
        eff_name = self._get_effsize_name(data)
        columns = [*columns, pcol, eff_name]
        data = data[columns]
        return data

    def _extract_data_friedman(self, data: pd.DataFrame, pcol: str, collapse_dof: bool) -> pd.DataFrame:
        rename_dict = {"ddof1": "df1", "ddof2": "df2"}
        data = data.rename(columns=rename_dict)
        columns = []
        if collapse_dof:
            dof_cols = ["df1", "df2"] if "df2" in data.columns else ["df1"]
            dofs = tuple(data[col].unique() for col in dof_cols)
            if any(len(d) != 1 for d in dofs):
                raise ValueError(f"Cannot collapse dof in table: dof are not unique! Got {dofs}.")
            dofs = [self._format_dof(dof[0]) for dof in dofs]
            if "F" in data.columns:
                f_col = MAP_LATEX_EXPORT["F_collapse"].format(*dofs)
                columns.append(f_col)
                data = data.rename(columns={"F": f_col})
                data = data.drop(columns=["df1", "df2", "df"], errors="ignore")
            else:
                q_col = MAP_LATEX_EXPORT["Q_collapse"].format(*dofs)
                columns.append(q_col)
                data = data.rename(columns={"Q": q_col})
                data = data.drop(columns=["df1", "df2", "df"], errors="ignore")
        else:
            if "df2" in data.columns:
                data["df"] = "{" + data["df1"].astype(str) + ", " + data["df2"].astype(str) + "}"
                data = data.drop(columns=["df1", "df2"])
            columns.append("df")
        columns = [*columns, pcol]
        data = data[columns]
        return data

    @staticmethod
    def _extract_data_mwu(data: pd.DataFrame, pcol: str) -> pd.DataFrame:
        columns = ["U-val", pcol, StatsPipeline._get_effsize_name(data)]
        data = data[columns]
        return data

    @staticmethod
    def _extract_data_wilcoxon(data: pd.DataFrame, pcol: str, show_a_b: bool) -> pd.DataFrame:
        columns = ["W-val", pcol, StatsPipeline._get_effsize_name(data)]
        data = data[columns]
        data = data.rename(columns={"W-val": "W"})
        if show_a_b and "A" in data.columns:
            data = StatsPipeline._remove_a_b(data)
        return data

    @staticmethod
    def _remove_a_b(data: pd.DataFrame) -> pd.DataFrame:
        data = data.set_index(["A", "B"], append=True)
        names_old = list(data.index.names)
        nlevels_old = len(names_old)
        names_new = names_old[: nlevels_old - 1] + names_old[nlevels_old:] + [names_old[nlevels_old - 1]]
        # reorder levels
        return data.reorder_levels(names_new)

    @staticmethod
    def _format_latex_table_index(
        data: pd.DataFrame, index_kws: dict[str, Any], show_a_b: Optional[bool] = False
    ):  # pylint:disable=too-many-branches
        index_italic = index_kws.get("index_italic", True)
        index_level_order = index_kws.get("index_level_order")
        index_value_order = index_kws.get("index_value_order")
        index_rename_map = index_kws.get("index_rename_map")
        index_level_names_tex = index_kws.get("index_level_names_tex")

        if isinstance(data.columns, pd.MultiIndex):
            data.columns.names = [None for _ in data.columns.names]
        else:
            data.columns.name = None

        if index_level_order is not None:
            data = data.reorder_levels(index_level_order)

        if index_value_order is not None:
            data = StatsPipeline._apply_index_value_order(data, index_value_order)

        if index_rename_map is not None:
            data = data.rename(index=index_rename_map)
        if index_level_names_tex is not None:
            if isinstance(index_level_names_tex, str):
                index_level_names_tex = [index_level_names_tex]
            if show_a_b and "A" not in index_level_names_tex:
                index_level_names_tex += ["A", "B"]
            data.index.names = index_level_names_tex

        if index_italic:
            data.index.names = [rf"\textit{{{s}}}" for s in data.index.names]
            data = data.T
            data = data.add_prefix(r"\textit{").add_suffix("}")
            data = data.T
        return data

    @staticmethod
    def _format_latex_table_column(data: pd.DataFrame, column_kws: dict[str, Any]):  # pylint:disable=too-many-branches
        column_level_order = column_kws.get("column_level_order")
        column_value_order = column_kws.get("column_value_order")
        column_rename_map = column_kws.get("column_rename_map")
        column_level_names_tex = column_kws.get("column_level_names_tex")

        if column_level_order is not None:
            data = data.reorder_levels(column_level_order, axis=1)

        if column_value_order is not None:
            data = StatsPipeline._apply_column_value_order(data, column_value_order)

        if column_rename_map is not None:
            data = data.rename(columns=column_rename_map)
        if column_level_names_tex is not None:
            if isinstance(column_level_names_tex, str):
                column_level_names_tex = [column_level_names_tex]
            data.index.names = column_level_names_tex

        data = data.add_prefix("{").add_suffix("}")
        return data

    @staticmethod
    def _apply_index_value_order(
        data: pd.DataFrame, index_value_order: Union[Sequence[str], dict[str, Sequence[str]]]
    ) -> pd.DataFrame:
        if isinstance(data.index, pd.MultiIndex):
            if isinstance(index_value_order, dict):
                for key, val in index_value_order.items():
                    data = data.reindex(val, level=key)
            else:
                raise TypeError(
                    "'index_value_order' must be a dictionary with index level names as keys and "
                    "index values as values."
                )
        elif isinstance(index_value_order, dict):
            data = data.reindex(index_value_order[data.index.name])
        else:
            data = data.reindex(index_value_order)
        return data

    @staticmethod
    def _apply_column_value_order(
        data: pd.DataFrame, column_value_order: Union[Sequence[str], dict[str, Sequence[str]]]
    ) -> pd.DataFrame:
        if isinstance(data.columns, pd.MultiIndex):
            if isinstance(column_value_order, dict):
                for key, val in column_value_order.items():
                    data = data.reindex(val, level=key, axis=1)
            else:
                raise TypeError(
                    "'column_value_order' must be a dictionary with column level names as keys and "
                    "column values as values."
                )
        elif isinstance(column_value_order, dict):
            data = data.reindex(column_value_order[data.index.name], axis=1)
        else:
            data = data.reindex(column_value_order, axis=1)
        return data

    @staticmethod
    def _format_latex_column_format(data: pd.DataFrame):
        column_format = "l" * data.index.nlevels + "||"
        if isinstance(data.columns, pd.MultiIndex):
            ncols = len(data.columns)
            ncols_last_level = len(data.columns.get_level_values(-1).unique())
            column_format += ("S" * ncols_last_level + "|") * (ncols // ncols_last_level)
            # remove the last "|"
            column_format = column_format[:-1]
        else:
            column_format += "S" * len(data.columns)
        return column_format

    @staticmethod
    def _apply_latex_code_correction(data_latex: str, si_table_format: str):
        data_latex = data_latex.replace(r"\textasciicircum ", "^")
        if si_table_format is not None:
            data_latex = re.sub(r"(\\begin\{tabular\})", r"\\sisetup{" + si_table_format + r"}\n\n\1", data_latex)
        return data_latex

    @staticmethod
    def _stats_to_index_set_index(data: pd.DataFrame) -> pd.DataFrame:
        if "Contrast" in data.columns:
            data = data.set_index("Contrast", append=True)
        if "Source" in data.columns:
            data = data.set_index("Source", append=True)
        if "A" in data.columns:
            data = data.set_index(["A", "B"], append=True)
        return data

    def _extract_data_latex_table(
        self, data: pd.DataFrame, stats_test: str, pcol: str, collapse_dof: bool, show_a_b: bool
    ) -> pd.DataFrame:
        if "anova" in stats_test:
            data = self._extract_data_anova(data, pcol, collapse_dof)
        if "friedman" in stats_test:
            data = self._extract_data_friedman(data, pcol, collapse_dof)
        if "T" in data.columns:
            data = self._extract_data_ttest(data, pcol, collapse_dof, show_a_b)
        if "U-val" in data.columns:
            data = self._extract_data_mwu(data, pcol)
        if "W-val" in data.columns:
            data = self._extract_data_wilcoxon(data, pcol, show_a_b)
        return data



"""Module to perform sleep/wake detection based on Actigraphy or IMU data."""

from biopsykit.sleep.sleep_wake_detection import algorithms
from biopsykit.sleep.sleep_wake_detection.sleep_wake_detection import SleepWakeDetection

__all__ = [
    "SleepWakeDetection",
    "algorithms",
]



import warnings
from typing import Literal

import pandas as pd
from typing_extensions import Self

__all__ = ["PepExtraction"]

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, BaseExtraction, CanHandleMissingEventsMixin
from biopsykit.utils.dtypes import (
    BPointDataFrame,
    HeartbeatSegmentationDataFrame,
    QPeakDataFrame,
    is_b_point_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_pep_result_dataframe,
    is_q_peak_dataframe,
)
from biopsykit.utils.exceptions import EventExtractionError

NEGATIVE_PEP_HANDLING = Literal["nan", "zero", "keep"]


class PepExtraction(BaseExtraction, CanHandleMissingEventsMixin):
    _action_methods = "extract"

    handle_negative_pep: NEGATIVE_PEP_HANDLING
    pep_results_: pd.DataFrame

    def __init__(
        self,
        handle_negative_pep: NEGATIVE_PEP_HANDLING = "nan",
        handle_missing_events: HANDLE_MISSING_EVENTS = "warn",
    ) -> None:
        self.handle_negative_pep = handle_negative_pep
        super().__init__(handle_missing_events=handle_missing_events)

    def extract(
        self,
        *,
        heartbeats: HeartbeatSegmentationDataFrame,
        q_peak_samples: QPeakDataFrame,
        b_point_samples: BPointDataFrame,
        sampling_rate_hz: float,
    ) -> Self:
        """Compute PEP from Q-peak samples and B-point locations.

        Args:
            heartbeats:
                Heartbeat locations as DataFrame
            q_peak_samples:
                ECG signal as DataFrame
            b_point_samples:
                ICG signal as DataFrame
            sampling_rate_hz:
                Sampling rate of the signals in Hz

        Returns
        -------
            self
        """
        is_heartbeat_segmentation_dataframe(heartbeats)
        is_b_point_dataframe(b_point_samples)
        is_q_peak_dataframe(q_peak_samples)
        self._check_valid_missing_handling()
        self._check_valid_handle_pep()

        # do something
        pep_results = pd.DataFrame(
            index=heartbeats.index,
            columns=[
                "heartbeat_start_sample",
                "heartbeat_end_sample",
                "r_peak_sample",
                "rr_interval_sample",
                "rr_interval_ms",
                "heart_rate_bpm",
                "q_peak_sample",
                "b_point_sample",
                "pep_sample",
                "pep_ms",
                "nan_reason",
            ],
        )

        if heartbeats.empty:
            missing_str = "No heartbeats found, no PEP can be extracted!"
            self.pep_results_ = pep_results
            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)
            return self

        pep_results = pep_results.assign(
            heartbeat_start_time=heartbeats["start_time"],
            heartbeat_start_sample=pd.to_numeric(heartbeats["start_sample"]),
            heartbeat_end_sample=pd.to_numeric(heartbeats["end_sample"]),
            r_peak_sample=pd.to_numeric(heartbeats["r_peak_sample"]),
            rr_interval_sample=pd.to_numeric(heartbeats["rr_interval_sample"]),
            rr_interval_ms=pd.to_numeric(heartbeats["rr_interval_sample"] / sampling_rate_hz * 1000),
            heart_rate_bpm=pd.to_numeric(60 / (heartbeats["rr_interval_sample"] / sampling_rate_hz)),
            q_peak_sample=pd.to_numeric(q_peak_samples["q_peak_sample"]),
            b_point_sample=pd.to_numeric(b_point_samples["b_point_sample"]),
            pep_sample=pd.to_numeric(b_point_samples["b_point_sample"] - q_peak_samples["q_peak_sample"]),
        )

        pep_results = pep_results.assign(
            pep_ms=pep_results["pep_sample"] / sampling_rate_hz * 1000,
        )
        pep_results = self._add_invalid_pep_reason(pep_results, q_peak_samples, b_point_samples)
        pep_results = pep_results.astype(
            {
                "heartbeat_start_sample": "Int64",
                "heartbeat_end_sample": "Int64",
                "r_peak_sample": "Int64",
                "rr_interval_sample": "Int64",
                "rr_interval_ms": "Float64",
                "heart_rate_bpm": "Float64",
                "q_peak_sample": "Int64",
                "b_point_sample": "Int64",
                "pep_sample": "Int64",
                "pep_ms": "Float64",
                "nan_reason": "object",
            }
        )

        is_pep_result_dataframe(pep_results)

        self.pep_results_ = pep_results

        return self

    def _add_invalid_pep_reason(
        self,
        pep_results: pd.DataFrame,
        q_peaks: QPeakDataFrame,
        b_points: BPointDataFrame,
    ) -> pd.DataFrame:
        # extract nan_reason from q_peak_samples and add to pep_results
        pep_results = pep_results.assign(nan_reason=q_peaks["nan_reason"])
        # TODO add option to store multiple nan_reasons in one column?
        # extract nan_reason from b_point_samples
        nan_reason_b_point = b_points["nan_reason"].loc[~b_points["nan_reason"].isna()]
        # add nan_reason to pep_results
        if not nan_reason_b_point.empty:
            pep_results.loc[nan_reason_b_point.index, "nan_reason"] = nan_reason_b_point

        neg_pep_idx = pep_results["pep_ms"] < 0
        if self.handle_negative_pep == "zero":
            pep_results.loc[neg_pep_idx, ["pep_sample", "pep_ms"]] = 0
            pep_results.loc[neg_pep_idx, "nan_reason"] = "negative_pep"
        elif self.handle_negative_pep == "nan":
            pep_results.loc[neg_pep_idx, ["pep_sample", "pep_ms"]] = pd.NA
            pep_results.loc[neg_pep_idx, "nan_reason"] = "negative_pep"

        return pep_results

    def _check_valid_handle_pep(self):
        if self.handle_negative_pep not in ["nan", "zero", "keep"]:
            raise ValueError(f"Invalid value for 'handle_negative_pep': {self.handle_negative_pep}")



"""Module containing utility functions for manipulating and processing questionnaire data."""

import re
import warnings
from collections.abc import Sequence
from inspect import getmembers, isfunction
from typing import Any, Literal, Optional, Union

import numpy as np
import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_is_dtype, _assert_len_list, _assert_value_range
from biopsykit.utils.dataframe_handling import wide_to_long as wide_to_long_utils

__all__ = [
    "bin_scale",
    "compute_scores",
    "convert_scale",
    "crop_scale",
    "find_cols",
    "get_supported_questionnaires",
    "invert",
    "to_idx",
    "wide_to_long",
    "zero_pad_columns",
]


def find_cols(
    data: pd.DataFrame,
    regex_str: Optional[str] = None,
    starts_with: Optional[str] = None,
    ends_with: Optional[str] = None,
    contains: Optional[str] = None,
    zero_pad_numbers: Optional[bool] = True,
) -> tuple[pd.DataFrame, Sequence[str]]:
    r"""Find columns in dataframe that match a specific pattern.

    This function is useful to find all columns that belong to a questionnaire. Column names can be filtered based on
    one (or a combination of) the following criteria:

    * ``starts_with``: columns have to start with the specified string
    * ``ends_with``: columns have to end with the specified string
    * ``contains``: columns have to contain the specified string

    Optionally, the item numbers in the matching column names can be zero-padded, if they are not already.

    .. note::
        If ``zero_pad_numbers`` is ``True`` then the column names returned by this function will be renamed and might
        thus not match the column names of the original dataframe. To solve this, make sure your orignal dataframe
        already has zero-padded columns (by manually renaming them) or convert column names using
        :func:`~biopsykit.questionnaires.utils.zero_pad_columns`.

    .. warning::
        Zero-padding using :func:`~biopsykit.questionnaires.utils.zero_pad_columns` assumes, by default, that numbers
        are *at the end* of column names. If you want to change that behavior
        (e.g., because the column names have string suffixes), you might need to apply zero-padding manually.



    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with columns to be filtered
    regex_str : str, optional
        regex string to extract column names. If this parameter is passed the other parameters (``starts_with``,
        ``ends_with``, ``contains``) will be ignored. Default: ``None``
    starts_with : str, optional
        string columns have to start with. Default: ``None``
    ends_with : str, optional
        string columns have to end with. Default: ``None``
    contains : str, optional
        string columns have to contain. Default: ``None``
    zero_pad_numbers : bool, optional
        whether to zero-pad numbers in column names. Default: ``True``

    Returns
    -------
    data_filt : :class:`~pandas.DataFrame`
        dataframe with filtered columns that match the specified pattern
    cols : :class:`~pandas.Index`
        columns that match the specified pattern

    Examples
    --------
    >>> import biopsykit as bp
    >>> import pandas as pd
    >>> # Option 1: has to start with "XX"
    >>> data = pd.DataFrame(columns=["XX_{}".format(i) for i in range(1, 11)])
    >>> df, cols = bp.questionnaires.utils.find_cols(data, starts_with="XX")
    >>> print(cols)
    >>> ["XX_01", "XX_02", ..., "XX_10"]
    >>> # Option 2: has to end with "Post"
    >>> data = pd.DataFrame(columns=["XX_1_Pre", "XX_2_Pre", "XX_3_Pre", "XX_1_Post", "XX_2_Post", "XX_3_Post"])
    >>> df, cols = bp.questionnaires.utils.find_cols(data, ends_with="Post")
    >>> print(cols)
    >>> ["XX_01_Post", "XX_02_Post", "XX_03_Post"]
    >>> # Option 3: has to start with "XX" and end with "Post"
    >>> data = pd.DataFrame(columns=["XX_1_Pre", "XX_2_Pre", "XX_3_Pre", "XX_1_Post", "XX_2_Post", "XX_3_Post",
     "YY_1_Pre", "YY_2_Pre", "YY_1_Post", "YY_2_Post"])
    >>> bp.questionnaires.utils.find_cols(data, starts_with="XX", ends_with="Post")
    >>> print(cols)
    >>> # WARNING: this will not zero-pad the questionnaire numbers!
    >>> ["XX_1_Post", "XX_2_Post", "XX_3_Post"]
    >>> # Option 4: pass custom regex string
    >>> data = pd.DataFrame(columns=["XX_1_Pre", "XX_2_Pre", "XX_3_Pre", "XX_1_Post", "XX_2_Post", "XX_3_Post",
     "YY_1_Pre", "YY_2_Pre", "YY_1_Post", "YY_2_Post"])
    >>> bp.questionnaires.utils.find_cols(data, regex_str=r"XX_\d+_\w+")
    >>> print(cols)
    >>> # here, zero-padding will be possible again
    >>> ["XX_01_Post", "XX_02_Post", "XX_03_Post"]
    >>> # Option 5: disable zero-padding
    >>> data = pd.DataFrame(columns=["XX_{}".format(i) for i in range(1, 11)])
    >>> df, cols = bp.questionnaires.utils.find_cols(data, starts_with="XX", zero_pad_numbers=False)
    >>> print(cols)
    >>> ["XX_1", "XX_2", ..., "XX_10"]

    """
    _assert_is_dtype(data, pd.DataFrame)
    data_filt = data.copy()

    if regex_str:
        data_filt = data_filt.filter(regex=regex_str)
    else:
        if starts_with:
            data_filt = data_filt.filter(regex="^" + starts_with)
        if ends_with:
            data_filt = data_filt.filter(regex=ends_with + "$")
        if contains:
            data_filt = data_filt.filter(regex=contains)

    if zero_pad_numbers:
        data_filt = zero_pad_columns(data_filt)
    cols = data_filt.columns

    return data_filt, cols


def zero_pad_columns(data: pd.DataFrame, inplace: Optional[bool] = False) -> Optional[pd.DataFrame]:
    r"""Add zero-padding to numbers at the **end** of column names in a dataframe.

    .. warning::
        By default, this function assumes that numbers are **at the end** of column names. If you need to change that
        behavior (e.g., because the column names have string suffixes), you might need to apply zero-padding manually.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with columns to zero-pad
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame` or ``None``
        dataframe with zero-padded columns or ``None`` if ``inplace`` is ``True``

    """
    _assert_is_dtype(data, pd.DataFrame)
    if not inplace:
        data = data.copy()

    nums = [re.findall(r"(\d+)$", c) for c in data.columns]
    nums = [c[0] if len(c) > 0 else "" for c in nums]
    if len(nums) == 0:
        return pd.DataFrame()
    zfill_num = max(*list(map(len, nums)), 2)

    data.columns = [re.sub(r"(\d+)$", lambda m: m.group(1).zfill(zfill_num), c) for c in data.columns]

    if inplace:
        return None
    return data


def to_idx(col_idxs: Union[np.array, Sequence[int]]) -> np.ndarray:
    """Convert questionnaire item indices into array indices.

    In questionnaires, items indices start at 1. To avoid confusion in the implementation of questionnaires
    (because array indices start at 0) all questionnaire indices in BioPsyKit also start at 1 and are converted to
    0-based indexing using this function.

    Parameters
    ----------
    col_idxs : list of int
        list of indices to convert to 0-based indexing

    Returns
    -------
    :class:`~numpy.ndarray`
        array with converted indices

    """
    return np.array(col_idxs) - 1


def invert(
    data: Union[pd.DataFrame, pd.Series],
    score_range: Sequence[int],
    cols: Optional[Union[np.array, Sequence[int], Sequence[str]]] = None,
    inplace: Optional[bool] = False,
) -> Optional[Union[pd.DataFrame, pd.Series]]:
    """Invert questionnaire scores.

    In many questionnaires some items need to be inverted (reversed) before sum scores can be computed. This function
    can be used to either invert a single column (Series), selected columns in a dataframe (by specifying columns
    in the ``cols`` parameter), or a complete dataframe.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        questionnaire data to invert
    score_range : list of int
        possible score range of the questionnaire items
    cols : list of str or list of int
        list of column names or column indices
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame` or ``None``
        dataframe with inverted columns or ``None`` if ``inplace`` is ``True``


    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is no dataframe or series
        if ``score_range`` does not have length 2
    :exc:`~biopsykit.utils.exceptions.ValueRangeError`
        if values in ``data`` are not in ``score_range``


    Examples
    --------
    >>> from biopsykit.questionnaires.utils import invert
    >>> data_in = pd.DataFrame({"A": [1, 2, 3, 1], "B": [4, 0, 1, 3], "C": [0, 3, 2, 3], "D": [0, 1, 2, 4]})
    >>> data_out = invert(data_in, score_range=[0, 4])
    >>> data_out["A"]
    >>> [3, 2, 1, 3]
    >>> data_out["B"]
    >>> [0, 4, 3, 1]
    >>> data_out["C"]
    >>> [4, 1, 2, 1]
    >>> data_out["D"]
    >>> [4, 3, 2, 0]
    >>> # Other score range
    >>> data_out = invert(data, score_range=[0, 5])
    >>> data_out["A"]
    >>> [3, 2, 1, 3]
    >>> data_out["B"]
    >>> [1, 5, 4, 2]
    >>> data_out["C"]
    >>> [5, 2, 3, 2]
    >>> data_out["D"]
    >>> [5, 4, 3, 1]
    >>> # Invert only specific columns
    >>> data_out = invert(data, score_range=[0, 4], cols=["A", "C"])
    >>> data_out["A"]
    >>> [3, 2, 1, 3]
    >>> data_out["B"]
    >>> [4, 0, 1, 3]
    >>> data_out["C"]
    >>> [4, 1, 2, 1]
    >>> data_out["D"]
    >>> [0, 1, 2, 4]

    """
    _assert_is_dtype(data, (pd.DataFrame, pd.Series))
    _assert_len_list(score_range, 2)

    if not inplace:
        data = data.copy()

    if isinstance(data, pd.DataFrame):
        data = _invert_dataframe(data, cols, score_range)
    else:
        _assert_value_range(data, score_range)
        data.iloc[:] = score_range[1] - data.iloc[:] + score_range[0]

    if inplace:
        return None
    return data


def _invert_dataframe(
    data: pd.DataFrame, cols: Union[Sequence[str], Sequence[int]], score_range: Sequence[int]
) -> pd.DataFrame:
    if cols is not None:
        if isinstance(cols[0], str):
            _assert_value_range(data[cols], score_range)
            data.loc[:, cols] = score_range[1] - data.loc[:, cols] + score_range[0]
        else:
            _assert_value_range(data.iloc[:, cols], score_range)
            data.iloc[:, cols] = score_range[1] - data.iloc[:, cols] + score_range[0]
    else:
        _assert_value_range(data, score_range)
        data.iloc[:, :] = score_range[1] - data.iloc[:, :] + score_range[0]

    return data


def _invert_subscales(
    data: pd.DataFrame,
    subscales: dict[str, Sequence[Union[str, int]]],
    idx_dict: dict[str, Sequence[int]],
    score_range: Sequence[int],
) -> pd.DataFrame:
    """Invert questionnaire scores from a dictionary of questionnaire subscales.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        questionnaire data to invert
    subscales : dict
        dictionary with subscale names (keys) and list of item indices or column names belonging to the
        individual subscales (values)
    idx_dict : dict
        dictionary with subscale names (keys) and indices of items that should be inverted (values)
    score_range : list of int
        possible score range of the questionnaire items

    Returns
    -------
    :class:`~pandas.DataFrame` or ``None``
        dataframe with inverted columns

    See Also
    --------
    invert : invert scores of questionnaire columns

    """
    _assert_is_dtype(data, pd.DataFrame)

    for scale_name, idxs in idx_dict.items():
        if scale_name in subscales:
            data = invert(data, cols=to_idx(np.array(subscales[scale_name])[idxs]), score_range=score_range)
    return data


def convert_scale(
    data: Union[pd.DataFrame, pd.Series],
    offset: int,
    cols: Optional[Union[pd.DataFrame, pd.Series]] = None,
    inplace: Optional[bool] = False,
) -> Optional[Union[pd.DataFrame, pd.Series]]:
    """Convert the score range of questionnaire items.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        questionnaire data to invert
    offset : int
        offset to add to questionnaire items
    cols : list of str or list of int
        list of column names or column indices
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame`, :class:`~pandas.Series`, or ``None``
        dataframe with converted columns or ``None`` if ``inplace`` is ``True``


    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``data`` is no dataframe or series

    Examples
    --------
    >>> from biopsykit.questionnaires.utils import convert_scale
    >>> data_in = pd.DataFrame({"A": [1, 2, 3, 1], "B": [4, 0, 1, 3], "C": [0, 3, 2, 3], "D": [0, 1, 2, 4]})
    >>> # convert data from range [0, 4] to range [1, 5]
    >>> data_out = convert_scale(data_in, offset=1)
    >>> data_out["A"]
    >>> [2, 3, 4, 2]
    >>> data_out["B"]
    >>> [5, 1, 2, 4]
    >>> data_out["C"]
    >>> [1, 4, 3, 4]
    >>> data_out["D"]
    >>> [1, 2, 3, 5]
    >>> data_in = pd.DataFrame({"A": [1, 2, 3, 1], "B": [4, 2, 1, 3], "C": [3, 3, 2, 3], "D": [4, 1, 2, 4]})
    >>> # convert data from range [1, 4] to range [0, 3]
    >>> data_out = convert_scale(data_in, offset=-1)
    >>> print(data_out)
    >>> # convert only specific columns
    >>> data_out = convert_scale(data_in, offset=-1, columns=["A", "C"])
    >>> print(data_out)

    """
    _assert_is_dtype(data, (pd.DataFrame, pd.Series))

    if not inplace:
        data = data.copy()

    if isinstance(data, pd.DataFrame):
        data = _convert_scale_dataframe(data, cols, offset)
    else:
        data.iloc[:] = data.iloc[:] + offset

    if inplace:
        return None
    return data


def _convert_scale_dataframe(
    data: pd.DataFrame, cols: Union[Sequence[int], Sequence[str]], offset: int
) -> pd.DataFrame:
    if cols is None:
        data.iloc[:, :] = data.iloc[:, :] + offset
    elif isinstance(cols[0], int):
        data.iloc[:, cols] = data.iloc[:, cols] + offset
    elif isinstance(cols[0], str):
        data.loc[:, cols] = data.loc[:, cols] + offset

    return data


def crop_scale(
    data: Union[pd.DataFrame, pd.Series],
    score_range: Sequence[int],
    set_nan: Optional[bool] = False,
    inplace: Optional[bool] = False,
) -> Optional[Union[pd.DataFrame, pd.Series]]:
    """Crop questionnaire scales, i.e., set values out of range to specific minimum and maximum values or to NaN.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        data to be cropped
    score_range : list of int
        possible score range of the questionnaire items. Values out of ``score_range`` are cropped.
    set_nan : bool, optional
        whether to set values out of range to NaN or to the values specified by ``score_range``. Default: ``False``
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame`, :class:`~pandas.Series`, or ``None``
        dataframe (or series) with cropped scales or ``None`` if ``inplace`` is ``True``

    """
    _assert_is_dtype(data, (pd.DataFrame, pd.Series))
    _assert_len_list(score_range, 2)

    if not inplace:
        data = data.copy()

    if set_nan:
        data.mask((data < score_range[0]) | (data > score_range[1]), inplace=True)  # noqa: PD002
    else:
        data.mask((data < score_range[0]), other=score_range[0], inplace=True)  # noqa: PD002
        data.mask((data > score_range[1]), other=score_range[1], inplace=True)  # noqa: PD002

    if inplace:
        return None
    return data


def bin_scale(
    data: Union[pd.DataFrame, pd.Series],
    bins: Union[int, Sequence[float], pd.IntervalIndex],
    cols: Optional[Union[Sequence[Union[int, str]], Union[int, str]]] = None,
    first_min: Optional[bool] = True,
    last_max: Optional[bool] = False,
    inplace: Optional[bool] = False,
    **kwargs,
) -> Optional[Union[pd.Series, pd.DataFrame]]:
    """Bin questionnaire scales.

    Questionnaire scales are binned using :func:`pandas.cut` according to the bins specified by ``bins``.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        data with scales to be binned
    bins : int or list of float or :class:`~pandas.IntervalIndex``
        The criteria to bin by. ``bins`` can have one of the following types:

        * ``int`` : Defines the number of equal-width bins in the range of ``data``. The range of ``data`` is extended
          by 0.1% on each side to include the minimum and maximum values of ``data``.
        * sequence of scalars : Defines the bin edges allowing for non-uniform width. No extension of the range of
          ``data`` is done.
        * :class:`~pandas.IntervalIndex` : Defines the exact bins to be used. Note that the ``IntervalIndex`` for
          ``bins`` must be non-overlapping.

    cols : list of str or list of int, optional
        column name/index (or list of such) to be binned or ``None`` to use all columns (or if ``data`` is a series).
        Default: ``None``
    first_min : bool, optional
        whether the minimum value should be added as the leftmost edge of the last bin or not.
        Only considered if ``bins`` is a list. Default: ``False``
    last_max : bool, optional
        whether the maximum value should be added as the rightmost edge of the last bin or not.
        Only considered if ``bins`` is a list. Default: ``False``
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``
    **kwargs
        additional parameters that are passed to :func:`pandas.cut`


    Returns
    -------
    :class:`~pandas.DataFrame`, :class:`~pandas.Series`, or ``None``
        dataframe (or series) with binned scales or ``None`` if ``inplace`` is ``True``


    See Also
    --------
    :func:`pandas.cut`
        Pandas method to bin values into discrete intervals.

    """
    _assert_is_dtype(data, (pd.Series, pd.DataFrame))

    if not inplace:
        data = data.copy()

    # set "labels" argument to False, but only if is wasn't specified by the user yet
    kwargs["labels"] = kwargs.get("labels", False)
    if isinstance(data, pd.Series):
        bins_c = _get_bins(data, bins, None, first_min, last_max)
        c = pd.cut(data.iloc[:], bins=bins_c, **kwargs)
        data.iloc[:] = c
        return data

    cols = _get_cols(data, cols)
    for col in cols:
        bins_c = _get_bins(data, bins, col, first_min, last_max)
        if isinstance(col, int):
            c = pd.cut(data.iloc[:, col], bins=bins_c, **kwargs)
            data.iloc[:, col] = c
        else:
            c = pd.cut(data.loc[:, col], bins=bins_c, **kwargs)
            data.loc[:, col] = c

    if inplace:
        return None
    return data


def wide_to_long(data: pd.DataFrame, quest_name: str, levels: Union[str, Sequence[str]]) -> pd.DataFrame:
    """Convert a dataframe wide-format into long-format.

    .. warning::
        This function is deprecated and will be removed in the future!
        Please use :func:`~biopsykit.utils.dataframe_handling.wide_to_long` instead.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        pandas DataFrame containing saliva data in wide-format, i.e. one column per saliva sample, one row per subject.
    quest_name : str
        questionnaire name, i.e., common name for each column to be converted into long-format.
    levels : str or list of str
        index levels of the resulting long-format dataframe.


    Returns
    -------
    :class:`~pandas.DataFrame`
        pandas DataFrame in long-format

    See Also
    --------
    :func:`~biopsykit.utils.dataframe_handling.wide_to_long`
        convert dataframe from wide to long format

    """
    warnings.warn(
        "'biopsykit.questionnaires.utils.wide_to_long()' is deprecated! "
        "Please update your code to use 'biopsykit.utils.dataframe_handling.wide_to_long()' in the future.",
        category=DeprecationWarning,
    )
    return wide_to_long_utils(data=data, stubname=quest_name, levels=levels)


def compute_scores(
    data: pd.DataFrame,
    quest_dict: dict[str, Union[Sequence[str], pd.Index]],
    quest_kwargs: Optional[dict[str, dict[str, Any]]] = None,
) -> pd.DataFrame:
    """Compute questionnaire scores from dataframe.

    This function can be used if multiple questionnaires from a dataframe should be computed at once. If the same
    questionnaire was assessed at multiple time points, these scores will be computed separately
    (see ``Notes`` and ``Examples``).
    The questionnaires (and the dataframe columns belonging to the questionnaires) are specified by ``quest_dict``.

    .. note::
        If questionnaires were collected at different time points (e.g., `pre` and `post`), which should all be
        computed, then the dictionary keys need to have the following format: "<questionnaire_name>-<time_point>".

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe containing questionnaire data
    quest_dict : dict
        dictionary with questionnaire names to be computed (keys) and columns of the questionnaires (values)
    quest_kwargs : dict
        dictionary with optional arguments to be passed to questionnaire functions. The dictionary is expected
        consist of questionnaire names (keys) and ``**kwargs`` dictionaries (values) with arguments per questionnaire

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with computed questionnaire scores

    Examples
    --------
    >>> from biopsykit.questionnaires.utils import compute_scores
    >>> quest_dict = {
    >>>     "PSS": ["PSS_{:02d}".format(i) for i in range(1, 11)], # PSS: one time point
    >>>     "PASA-pre": ["PASA_{:02d}_T0".format(i) for i in range(1, 17)], # PASA: two time points (pre and post)
    >>>     "PASA-post": ["PASA_{:02d}_T1".format(i) for i in range(1, 17)], # PASA: two time points (pre and post)
    >>> }
    >>> compute_scores(data, quest_dict)

    """
    from biopsykit.questionnaires import questionnaires  # pylint:disable=import-outside-toplevel

    _assert_is_dtype(data, pd.DataFrame)

    df_scores = pd.DataFrame(index=data.index)
    quest_funcs = dict(getmembers(questionnaires, isfunction))
    if quest_kwargs is None:
        quest_kwargs = {}

    for score, columns in quest_dict.items():
        score_orig = score
        score = score.lower()  # noqa: PLW2901
        suffix = None
        if "-" in score:
            score_split = score.split("-")
            score = score_split[0]  # noqa: PLW2901
            suffix = score_split[1]

        if score not in quest_funcs:
            raise ValueError(
                f"Unknown questionnaire '{score}'! Call "
                "'biopsykit.questionnaires.utils.get_supported_questionnaires()' "
                "to get a list of all supported questionnaires."
            )
        kwargs = quest_kwargs.get(score_orig, {})

        try:
            df = quest_funcs[score](data[columns], **kwargs)
        except TypeError as e:
            raise TypeError(
                f"Error computing questionnaire '{score}'. The computation failed with the following "
                f"error: \n\n{e!s}."
            ) from e

        if suffix is not None:
            df.columns = [f"{col}_{suffix}" for col in df.columns]
        df_scores = df_scores.join(df)

    return df_scores


def get_supported_questionnaires() -> dict[str, str]:
    """List all supported (i.e., implemented) questionnaires.

    Returns
    -------
    dict
        dictionary with questionnaire names (keys) and description (values)

    """
    from biopsykit.questionnaires import questionnaires  # pylint:disable=import-outside-toplevel

    funcs = dict(getmembers(questionnaires, isfunction))
    quests = {}
    for key, value in funcs.items():
        if key.startswith("_"):
            continue
        summary = value.__doc__.split("\n")[0]
        summary = re.findall(r"\*\*(.*)\*\*.", summary)
        if len(summary) == 0:
            continue
        quests[key] = summary[0]

    return quests


def _compute_questionnaire_subscales(
    data: pd.DataFrame,
    score_name: str,
    subscales: dict[str, Sequence[Union[str, int]]],
    agg_type: Optional[Literal["sum", "mean"]] = "sum",
) -> dict[str, pd.Series]:
    """Compute questionnaire subscales (helper function).

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe containing questionnaire data
    score_name : str
        name of the questionnaire
    subscales : dict
        dictionary with subscales to be computed. Keys are subscale names, values are the indices of the items
        belonging to the subscales
    agg_type : str
        whether to compute a ``sum`` or a ``mean`` score. Default: ``sum``

    Returns
    -------
    dict
        dictionary with computed subscales

    """
    _assert_is_dtype(data, pd.DataFrame)

    out = {}
    for key, items in subscales.items():
        if all(np.issubdtype(type(i), np.integer) for i in items):
            # assume column indices, starting at 1 (-> convert to 0-indexed indices first)
            score = _compute_questionnaire_scores_int(data, items, agg_type)
        elif all(isinstance(i, str) for i in items):
            # assume column names
            score = _compute_questionnaire_scores_str(data, items, agg_type)
        else:
            raise ValueError(
                "Subscale columns are either expected as column names (list of strings) or "
                "column indices (list of integers)!"
            )

        out[f"{score_name}_{key}"] = score

    return out


def _compute_questionnaire_scores_int(data: pd.DataFrame, items: Sequence[int], agg_type: str):
    if agg_type == "sum":
        return data.iloc[:, to_idx(items)].sum(axis=1)
    return data.iloc[:, to_idx(items)].mean(axis=1)


def _compute_questionnaire_scores_str(data: pd.DataFrame, items: Sequence[str], agg_type: str):
    if agg_type == "sum":
        return data.loc[:, items].sum(axis=1)
    return data.loc[:, items].mean(axis=1)


def _get_cols(
    data: pd.DataFrame, cols: Optional[Union[Sequence[Union[int, str]], Union[int, str]]] = None
) -> Sequence[Union[str, int]]:
    if isinstance(cols, int):
        cols = [cols]
    if isinstance(cols, str):
        cols = [cols]
    if cols is None:
        cols = list(data.columns)
    return cols


def _get_bins(
    data: Union[pd.DataFrame, pd.Series],
    bins: Union[int, Sequence[float]],
    col: Optional[Union[int, str]] = None,
    first_min: Optional[bool] = False,
    last_max: Optional[bool] = False,
) -> Union[int, Sequence[float]]:
    if isinstance(bins, (int, pd.IntervalIndex)):
        return bins

    # ensure list
    bins = list(bins)

    if first_min:
        min_val = _bin_scale_get_min_val(data, col)
        if min_val < min(bins):
            bins = [min_val - 0.01, *bins]

    if last_max:
        max_val = _bin_scale_get_max_val(data, col)
        if max_val > max(bins):
            bins = [*bins, max_val + 0.01]

    return bins


def _bin_scale_get_min_val(data: Union[pd.DataFrame, pd.Series], col: Union[int, str]) -> float:
    if isinstance(col, int):
        return data.iloc[:, col].min()
    if isinstance(col, str):
        return data[col].min()
    return data.min()


def _bin_scale_get_max_val(data: Union[pd.DataFrame, pd.Series], col: Union[int, str]) -> float:
    if isinstance(col, int):
        return data.iloc[:, col].max()
    if isinstance(col, str):
        return data[col].max()
    return data.max()



"""Sleep/Wake detection using the *Cole/Kripke Algorithm*."""

import numpy as np
import pandas as pd

from biopsykit.sleep.sleep_wake_detection.algorithms._base import _SleepWakeBase
from biopsykit.sleep.sleep_wake_detection.utils import rescore
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.array_handling import sanitize_input_1d
from biopsykit.utils.dtypes import SleepWakeDataFrame, _SleepWakeDataFrame


class ColeKripke(_SleepWakeBase):
    """Class representing the *Cole/Kripke Algorithm* for sleep/wake detection."""

    def __init__(self, **kwargs):  # pylint: disable=useless-super-delegation
        """Class representing the *Cole/Kripke Algorithm* for sleep/wake detection.

        The *Cole/Kripke Algorithm* runs sleep wake detection on epoch level activity data. Epochs are 1 minute long and
        activity is represented by an activity index which comes from Actigraph data or from raw acceleration data
        converted into activity index data.

        Parameters
        ----------
        epoch_length : int
            epoch length in seconds. Epoch lengths are usually 10, 30, or 60 seconds.
        scale_factor : float
            scale factor to use for the predictions (default corresponds to scale factor optimized for use with
            the activity index, if other activity measures are desired the scale factor can be modified or optimized.)
            The recommended range for the scale factor is between 0.1 and 0.25 depending on the sensitivity to activity
            desired, and possibly the population being observed. According to the paper by Cole and Kripke,
            the scale factor depends on the epoch length. See the paper for more details.

        References
        ----------
        Cole, R. J., Kripke, D. F., Gruen, W., Mullaney, D. J., & Gillin, J. C. (1992). Automatic Sleep/Wake
        Identification From Wrist Activity. *Sleep*, 15(5), 461-469. https://doi.org/10.1093/sleep/15.5.461

        """
        super().__init__(**kwargs)

    def fit(self, data: arr_t, **kwargs):  # noqa: ARG002
        """Fit sleep/wake detection algorithm to input data.

        .. note::
            Algorithms that do not have to (re)fit a ML model before sleep/wake prediction, such as rule-based
            algorithms, will internally bypass this method as the ``fit`` step is not needed.

        Parameters
        ----------
        data : array_like
            input data

        """
        return

    def predict(self, data: arr_t, **kwargs) -> SleepWakeDataFrame:
        """Apply sleep/wake prediction on activity index values.

        Parameters
        ----------
        data : array_like
            array with activity index values
        **kwargs :
            additional arguments to be passed to the algorithm for prediction, such as:

            * ``rescore_data`` (``bool``):
              ``True`` to apply Webster's rescoring rules to the sleep/wake predictions, ``False`` otherwise.
              Default: ``True``
            * ``epoch_length`` (``int``):
              activity data epoch lengths in seconds, i.e. Epoch lengths are usually 30 or 60 seconds.
              Default: 60

        Returns
        -------
        :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`
            dataframe with sleep/wake predictions

        """
        index = None
        rescore_data: bool = kwargs.get("rescore_data", True)

        if isinstance(data, pd.DataFrame):
            index = data.index
            data = sanitize_input_1d(data)

        # ensure numpy
        # problem in my view: with np.convolve and "same" we don't make the "unsymmetrical" convolution...
        # ( formula is with A-4 and A+2)
        # possible solution: np.valid and append 4 zeros in front and 2 zeros at the back or like implemented
        scores = np.convolve(data, self._get_kernel(), "same")

        scores[scores >= 1] = 99  # wake = 0
        scores[scores < 1] = 1  # sleep = 1
        scores[scores == 99] = 0  # wake = 0       #changed to 1 according to paper

        if rescore_data:
            scores = rescore(scores, epoch_length=self.epoch_length)

        if index is not None:
            scores = pd.DataFrame(scores, index=index, columns=["sleep_wake"])

        return _SleepWakeDataFrame(scores)

    def _get_kernel(self):
        """Get convolution kernel for a given epoch length."""
        scale_map = {
            10: 0.00001,
            30: 0.0001,
            60: 0.001,
        }
        coeff_map = {
            10: [550, 378, 413, 699, 1736, 287, 309, 0, 0],
            30: [50, 30, 14, 28, 121, 8, 50, 0, 0],
            60: [106, 54, 58, 76, 230, 74, 67, 0, 0],
        }
        sf = np.array(scale_map[self.epoch_length])
        return sf * np.array(coeff_map[self.epoch_length])



import pandas as pd
from tpcp import Algorithm

__all__ = ["BaseHeartbeatSegmentation"]

from biopsykit.utils.dtypes import EcgRawDataFrame


class BaseHeartbeatSegmentation(Algorithm):
    """Base class for all heartbeat segmentation algorithms."""

    _action_methods = "extract"

    # result
    heartbeat_list_: pd.DataFrame

    def extract(
        self,
        *,
        ecg: EcgRawDataFrame,
        sampling_rate_hz: float,
    ):
        """Segment ECG signal into heartbeats.

        Parameters
        ----------
        ecg : EcgRawDataFrame
            ECG data.
        sampling_rate_hz : float
            Sampling rate of the ECG data in Hz.

        """
        raise NotImplementedError("Method 'extract' must be implemented in subclass.")



"""Module with functions and classes to help with model selection for classification."""

from biopsykit.classification.model_selection.nested_cv import nested_cv_param_search
from biopsykit.classification.model_selection.sklearn_pipeline_permuter import SklearnPipelinePermuter

__all__ = ["SklearnPipelinePermuter", "nested_cv_param_search"]



from typing import Optional

from biopsykit.signals._base_extraction import BaseExtraction

__all__ = ["BaseCPointExtraction"]

from biopsykit.utils.dtypes import CPointDataFrame, HeartbeatSegmentationDataFrame, IcgRawDataFrame


class BaseCPointExtraction(BaseExtraction):
    """Base class for C-point extraction algorithms."""

    points_: CPointDataFrame

    def extract(
        self, *, icg: IcgRawDataFrame, heartbeats: HeartbeatSegmentationDataFrame, sampling_rate_hz: Optional[float]
    ):
        raise NotImplementedError("This is an abstract method and needs to be implemented in a subclass.")



"""Functions to analyze classification results."""

from collections.abc import Sequence
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
import seaborn as sns
from fau_colors import cmaps
from matplotlib import pyplot as plt
from matplotlib.cm import ColormapRegistry
from matplotlib.colors import ListedColormap
from sklearn.metrics import ConfusionMatrixDisplay

from biopsykit.classification.model_selection import SklearnPipelinePermuter
from biopsykit.classification.utils import prepare_df_sklearn

pipeline_step_map = {
    "pipeline_scaler": "Scaler",
    "pipeline_reduce_dim": r"\makecell[lc]{Feature\\ Selection}",
    "pipeline_clf": "Classifier",
}

metric_map = {
    "accuracy": r"\makecell{Accuracy [\%]}",
    "f1": r"\makecell{F1-score [\%]}",
    "precision": r"\makecell{Precision [\%]}",
    "recall": r"\makecell{Recall [\%]}",
    "auc": r"\makecell{AUC [\%]}",
    "sensitivity": r"\makecell{Sensitivity [\%]}",
    "specificity": r"\makecell{Specificity [\%]}",
}

clf_map = {
    "MinMaxScaler": "Min-Max",
    "StandardScaler": "Standard",
    "SelectKBest": "SkB",
    "RFE": "RFE",
    "GaussianNB": "NB",
    "KNeighborsClassifier": "kNN",
    "DecisionTreeClassifier": "DT",
    "SVC": "SVM",
    "RandomForestClassifier": "RF",
    "MLPClassifier": "MLP",
    "AdaBoostClassifier": "Ada",
}


def predictions_as_df(
    pipeline_permuter: SklearnPipelinePermuter,
    data: pd.DataFrame,
    pipeline: tuple[str],
    label_mapping: Optional[dict[str, str]] = None,
    index_col: Optional[str] = None,
) -> pd.DataFrame:
    """Get predictions from a specified pipeline and merge them with the index of the input dataframe.

    Parameters
    ----------
    pipeline_permuter : :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
        :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter` instance
    data : :class:`~pandas.DataFrame`
        input data
    pipeline : tuple
        pipeline to get predictions from
    label_mapping : dict, optional
        mapping of labels to rename labels in the output dataframe or ``None`` to keep original labels.
        Default: ``None``
    index_col : str, optional
        name of the index column to merge the predictions with. If ``data`` has a multi-index,
        the first level is used unless ``index_col`` is specified.
        Default: ``None``

    Returns
    -------
    :class:`~pandas.DataFrame`
        predictions as dataframe

    """
    metric_summary = pipeline_permuter.metric_summary()
    label_cols = ["true_labels", "predicted_labels"]
    predictions = metric_summary[label_cols].explode(label_cols).loc[pipeline]

    if isinstance(data.index, pd.MultiIndex):
        if index_col is None:
            index_col = data.index.names[0]
        index_vals = data.index.get_level_values(index_col)
    else:
        index_vals = data.index

    predictions.index = index_vals
    if label_mapping:
        predictions = predictions.replace(label_mapping)
    return predictions


def predict_proba_from_estimator(
    pipeline_permuter: SklearnPipelinePermuter,
    data: pd.DataFrame,
    pipeline: tuple[str],
    label_col: Optional[str] = "label",
    column_names: Optional[dict[str, str]] = None,
) -> pd.DataFrame:
    """Get predictions as probabilities from a specified pipeline and merge them with the index of the input dataframe.

    Parameters
    ----------
    pipeline_permuter : :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter`
        :class:`~biopsykit.classification.model_selection.SklearnPipelinePermuter` instance
    data : :class:`~pandas.DataFrame`
        input data
    pipeline : tuple
        pipeline to get predictions from
    label_col : str, optional
        name of the label column in the input dataframe. Default: ``"label"``
    column_names : dict, optional
        mapping of column names to rename columns in the output dataframe or ``None`` to keep original column names.
        Default: ``None``

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with predictions as probabilities

    """
    metric_summary = pipeline_permuter.metric_summary()
    best_pipeline = pipeline_permuter.best_estimator_summary()
    best_pipeline = best_pipeline.loc[pipeline].iloc[0].pipeline

    test_indices = metric_summary.loc[pipeline]["test_indices_folds"]
    test_indices_flat = list(metric_summary.loc[pipeline]["test_indices"])

    x, y, _, _ = prepare_df_sklearn(data, label_col=label_col, print_summary=False)

    label_order = best_pipeline[0].classes_

    predict_proba_results = []
    predict_proba_labels = []

    for i, test_idx in enumerate(test_indices):
        test_idx_list = list(test_idx)
        pipeline_fold = best_pipeline[i]
        predict_proba_results.append(pipeline_fold.predict_proba(x[test_idx_list]))
        predict_proba_labels.append(y[test_idx_list])

    results_proba = pd.DataFrame(
        np.concatenate(predict_proba_results),
        columns=label_order,
        index=data.index[test_indices_flat],
    )
    if column_names is not None:
        results_proba = results_proba.rename(columns=column_names)
    results_proba = results_proba.sort_index()
    results_proba = results_proba.round(4)

    return results_proba


def _conf_matrix_from_proba_df(
    data: pd.DataFrame, label_col: Optional[str] = "label", label_order: Optional[Sequence[str]] = None
) -> pd.DataFrame:
    """Get confusion matrix from a dataframe with predictions as probabilities.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with predictions as probabilities
    label_col : str, optional
        name of the label column in the input dataframe. Default: ``"label"``
    label_order : list, optional
        order of labels in the confusion matrix or ``None`` to use the order of the labels in the input dataframe.
        Default: ``None``

    Returns
    -------
    :class:`~pandas.DataFrame`
        confusion matrix as dataframe

    """
    if label_order is None:
        label_order = list(data.columns)
    conf_matrix_proba = data.groupby(label_col).mean()
    conf_matrix_proba = conf_matrix_proba.reindex(label_order, axis=0).reindex(label_order, axis=1)
    return conf_matrix_proba


def _register_fau_r():
    fau_r = sns.color_palette(cmaps.fau)
    fau_r.reverse()
    fau_r = ListedColormap(fau_r, "fau_r")
    if "fau_r" not in plt.colormaps():
        colormap_registry = ColormapRegistry()
        colormap_registry.register(cmap=fau_r, name="fau_r")


def plot_conf_matrix(
    predictions: pd.DataFrame,
    labels: Sequence[str],
    label_name: Optional[str] = "label",
    conf_matrix_kwargs: Optional[dict[str, Any]] = None,
    **kwargs,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot confusion matrix from predictions.

    Parameters
    ----------
    predictions : :class:`~pandas.DataFrame`
        dataframe with predictions
    labels : list, dict, optional
        list of labels to use in the confusion matrix or dictionary with label names in the data frame as key and the
        corresponding label names to use in the confusion matrix as value.
        Default: ``None`` to use the labels in the data frame in the order they appear
    label_name : str, optional
        name of the 'label' in the axis titles. Default: "label" to yield "True label" and "Predicted label"
    conf_matrix_kwargs : dict, optional
        additional keyword arguments to pass to :func:`~sklearn.metrics.ConfusionMatrixDisplay.from_predictions`
    **kwargs
        additional keyword arguments to pass to :func:`plt.subplots`

    """
    if conf_matrix_kwargs is None:
        conf_matrix_kwargs = {}
    # check if ax is given
    ax = kwargs.get("ax")
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = ax.get_figure()

    predictions = predictions.copy()
    if isinstance(labels, dict):
        # only replace in true_labels and predicted_labels columns
        predictions[["true_labels", "predicted_labels"]] = predictions[["true_labels", "predicted_labels"]].replace(
            labels
        )
        labels = list(labels.values())

    if not conf_matrix_kwargs.get("cmap", None):
        # check if fau_r is registered as colormap
        if "fau_r" not in plt.colormaps():
            _register_fau_r()
        conf_matrix_kwargs["cmap"] = "fau_r"
    conf_matrix_kwargs.setdefault("colorbar", False)

    ConfusionMatrixDisplay.from_predictions(
        predictions["true_labels"],
        predictions["predicted_labels"],
        labels=labels,
        ax=ax,
        **conf_matrix_kwargs,
    )
    if kwargs.get("despine", True):
        for spine in ["top", "bottom", "left", "right"]:
            ax.spines[spine].set_color("None")

    ax.set_ylabel(f"True {label_name}")
    ax.set_xlabel(f"Predicted {label_name}")

    return fig, ax


def plot_conf_matrix_proba(
    predictions: pd.DataFrame,
    labels: Sequence[str],
    label_col: Optional[str] = "label",
    label_name: Optional[str] = "label",
    **kwargs,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot confusion matrix from prediction probabilities.

    Parameters
    ----------
    predictions : :class:`~pandas.DataFrame`
        dataframe with predictions as probabilities
    labels : list
        list of labels
    label_col : str, optional
        name of the label column in the input dataframe. Default: ``"label"``
    label_name : str, optional
        name of the 'label' in the axis titles. Default: "label" to yield "True label" and "Predicted label"
    **kwargs
        additional keyword arguments to pass to :func:`plt.subplots`

    """
    # check if ax is given
    ax = kwargs.get("ax")
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = ax.get_figure()

    # check if fau_r is registered as colormap
    if "fau_r" not in plt.colormaps():
        _register_fau_r()

    conf_matrix_proba = _conf_matrix_from_proba_df(predictions, label_col=label_col, label_order=labels)

    sns.heatmap(conf_matrix_proba, cmap="fau_r", annot=True, cbar=False, square=True, ax=ax)
    ax.set_ylabel(f"True {label_name}")
    ax.set_xlabel(f"Predicted {label_name}")

    return fig, ax


def metric_summary_to_latex(
    permuter_or_df: Union[SklearnPipelinePermuter, pd.DataFrame],
    metrics: Optional[Sequence[str]] = None,
    pipeline_steps: Optional[Sequence[str]] = None,
    si_table_format: Optional[str] = None,
    highlight_best: Optional[str] = None,
    **kwargs,
) -> str:
    """Return a latex table with the performance metrics of the pipeline combinations.

    Notes
    -----
    This method is a legacy method that is kept for backwards compatibility with older pickled instances of the
    :class:`SklearnPipelinePermuter` class. It is recommended to use the :meth:`SklearnPipelinePermuter.metric_summary`
    method instead.

    See Also
    --------
    :meth:`SklearnPipelinePermuter.metric_summary_to_latex`


    Parameters
    ----------
    permuter_or_df : :class:`SklearnPipelinePermuter` or :class:`~pandas.DataFrame`
        :class:`SklearnPipelinePermuter` instance or dataframe with performance metrics.
    metrics : list of str, optional
        list of metrics to include in the table or ``None`` to use all available metrics in the dataframe.
        Default: ``None``
    pipeline_steps : list of str, optional
        list of pipeline steps to include in the table index or ``None`` to show all available pipeline steps
        as table index. Default: ``None``
    si_table_format : str, optional
        table format for the ``siunitx`` package or ``None`` to use the default format. Default: ``None``
    highlight_best : bool or str, optional
        Whether to highlight the pipeline with the best value in each column or not.
        *  If ``highlight_best`` is a boolean, the best pipeline is highlighted in each column.
        *  If ``highlight_best`` is a string, the best pipeline is highlighted in the column with the name
    **kwargs
        additional keyword arguments passed to :func:`~pandas.DataFrame.to_latex`

    """
    dummy_permuter = SklearnPipelinePermuter()
    if isinstance(permuter_or_df, SklearnPipelinePermuter):
        metric_summary = permuter_or_df.metric_summary
    else:
        metric_summary = permuter_or_df

    return dummy_permuter.metric_summary_to_latex(
        metric_summary,
        metrics=metrics,
        pipeline_steps=pipeline_steps,
        si_table_format=si_table_format,
        highlight_best=highlight_best,
        **kwargs,
    )



"""Module for ECG data analysis and visualization."""

from biopsykit.signals.ecg import event_extraction, plotting, preprocessing, segmentation
from biopsykit.signals.ecg.ecg import EcgProcessor

__all__ = ["EcgProcessor", "event_extraction", "plotting", "preprocessing", "segmentation"]



"""Module with utility functions for machine learning and classification applications."""

from copy import deepcopy
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline

__all__ = ["factorize_subject_id", "prepare_df_sklearn", "split_train_test", "strip_df", "strip_labels"]


class _PipelineWrapper:
    def __init__(self, pipeline: Pipeline):
        self.pipeline = pipeline

    def __str__(self):
        return str(self.pipeline)

    def __repr__(self):
        return repr(self.pipeline)


# TODO check if works?
# def __getitem__(self, item):
#     return self.pipeline[item]
#
# def __len__(self):
#     return len(self.pipeline)


def strip_df(data: pd.DataFrame) -> np.ndarray:
    """Strip dataframe from all index levels to only contain values.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input dataframe

    Returns
    -------
    :class:`~numpy.ndarray`
        array of stripped dataframe without index

    """
    return np.array(data.reset_index(drop=True).values)


def strip_labels(data: Union[pd.DataFrame, pd.Series], label_col: Optional[str] = None) -> np.ndarray:
    """Strip labels from dataframe index.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        input data
    label_col : str, optional
        name of index level containing class labels or ``None`` to use default column name ("label").
        Default: ``None``

    Returns
    -------
    :class:`~numpy.ndarray`
        array with labels

    """
    # TODO change to dataframe column
    if label_col is None:
        label_col = "label"
    if isinstance(data, pd.DataFrame):
        data = data.index.get_level_values(label_col)
    return np.array(data)


def factorize_subject_id(
    data: Union[pd.Series, pd.DataFrame], subject_col: Optional[str] = None
) -> tuple[np.ndarray, np.ndarray]:
    """Factorize subject IDs, i.e., encode them as an enumerated type or categorical variable.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        input data
    subject_col : str, optional
        name of index level containing subject IDs or ``None`` to use default column name ("subject").
        Default: ``None``

    Returns
    -------
    groups : :class:`~numpy.ndarray`
        A numpy array with factorized subject IDs. They also serve as indexer for ``keys``.
    keys : :class:`~numpy.ndarray`
        The unique subject ID values.

    """
    if subject_col is None:
        subject_col = "subject"

    if isinstance(data, pd.DataFrame):
        data = data.index.get_level_values(subject_col)
    groups, keys = pd.factorize(data)
    return groups, keys


def prepare_df_sklearn(
    data: pd.DataFrame,
    label_col: Optional[str] = None,
    subject_col: Optional[str] = None,
    print_summary: Optional[bool] = False,
) -> tuple[np.ndarray, ...]:
    """Prepare a dataframe for usage in sklearn functions and return the single components of the dataframe.

    This function performs the following steps:

    * Strip dataframe from all index levels and return an array that only contains values
      (using :func:`~biopsykit.classification.utils.strip_df`)
    * Extract labels from dataframe (using :func:`~biopsykit.classification.utils.strip_labels`)
    * Factorize subject IDs so that each subject ID has an unique number
      (using :func:`~biopsykit.classification.utils.factorize_subject_id`)


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        Input data as pandas dataframe
    label_col : str, optional
        name of index level containing class labels or ``None`` to use default column name ("label").
        Default: ``None``
    subject_col : str, optional
        name of index level containing subject IDs or ``None`` to use default column name ("subject").
        Default: ``None``
    print_summary : bool, optional
        ``True`` to print a summary of the shape of the data and label arrays, the number of groups and the class
        prevalence of all classes, ``False`` otherwise.
        Default: ``False``

    Returns
    -------
    X : array-like of shape (`n_samples`, `n_features`)
        Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.
    y_data : array-like of shape (`n_samples`,)
        Target relative to ``X``, i.e. class labels.
    groups : array-like of shape (`n_samples`,)
        Factorized subject IDs
    group_keys : array-like of shape (`n_samples`,)
        Subject IDs

    """
    x_data = strip_df(data)
    y_data = strip_labels(data, label_col)
    groups, group_keys = factorize_subject_id(data, subject_col)

    if print_summary:
        print(
            f"Shape of X: {x_data.shape}; shape of y: {y_data.shape}; "
            f"number of groups: {len(group_keys)}, class prevalence: {np.unique(y_data, return_counts=True)[1]}"
        )

    return x_data, y_data, groups, group_keys


def split_train_test(
    X: np.ndarray,  # noqa: N803
    y: np.ndarray,
    train: np.ndarray,
    test: np.ndarray,
    groups: Optional[np.ndarray] = None,
) -> tuple[np.ndarray, ...]:
    """Split data into train and test set.

    Parameters
    ----------
    X : array-like of shape (`n_samples`, `n_features`)
        Data to be split, where `n_samples` is the number of samples and `n_features` is the number of features.
    y : array-like of shape (`n_samples`,)
        Target relative to ``x_data``, i.e. class labels.
    train : :class:`~numpy.ndarray`
        The training set indices for that split
    test : :class:`~numpy.ndarray`
        The test set indices for that split
    groups : array-like of shape (`n_samples`,), optional
        Group labels for the samples used while splitting the dataset into train/test set or ``None`` if group labels
        should not be considered for splitting.
        Default: ``None``


    Returns
    -------
    X_train: :class:`~numpy.ndarray`
        Training data
    X_test: :class:`~numpy.ndarray`
        Test data
    y_train: :class:`~numpy.ndarray`
        Targets of training data
    y_test: :class:`~numpy.ndarray`
        Targets of test data
    group_train: :class:`~numpy.ndarray`
        Group labels of training data (only available if ``groups`` is not ``None``)
    group_test: :class:`~numpy.ndarray`
        Group labels of test data (only available if ``groups`` is not ``None``)

    """
    X_train, X_test = X[train], X[test]  # noqa: N806
    y_train, y_test = y[train], y[test]

    if groups is None:
        return X_train, X_test, y_train, y_test

    groups_train = groups[train]
    groups_test = groups[test]
    return X_train, X_test, y_train, y_test, groups_train, groups_test


def merge_nested_dicts(dict1: dict, dict2: dict) -> dict:
    """Merge two nested dictionaries.

    Parameters
    ----------
    dict1 : dict
        First dictionary to merge
    dict2 : dict
        Second dictionary to merge

    Returns
    -------
    dict
        Merged dictionary

    """
    dict1 = deepcopy(dict1)
    return _merge_nested_dicts(dict1, dict2)


def _merge_nested_dicts(dict1: dict, dict2: dict) -> dict:
    for key, value in dict2.items():
        if isinstance(value, dict) and key in dict1:
            _merge_nested_dicts(dict1[key], value)
            # check if value is list
        elif isinstance(value, list) and key in dict1:
            dict1[key] = value if key not in dict1 else dict1[key] + value
            list_of_dicts = deepcopy(dict1[key])
            merged_dict = {}

            for d in list_of_dicts:
                for k, v in d.items():
                    # Use set to avoid duplicates, then convert it back to a list
                    merged_dict[k] = list(set(merged_dict.get(k, []) + list(v)))

            # Convert the merged result back to dictionaries
            result = [dict(merged_dict)]
            dict1[key] = result
        elif key not in dict1:
            dict1[key] = value
    return dict1



"""Sleep/Wake detection using the *Sazonov Algorithm*."""

import numpy as np
import pandas as pd
from scipy.special import expit

from biopsykit.sleep.sleep_wake_detection.algorithms._base import _SleepWakeBase
from biopsykit.sleep.sleep_wake_detection.utils import rescore
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.array_handling import sanitize_input_1d, sliding_window
from biopsykit.utils.dtypes import SleepWakeDataFrame, _SleepWakeDataFrame


class Sazonov(_SleepWakeBase):
    """Class representing the *Sazonov Algorithm* for sleep/wake detection based on activity counts."""

    def __init__(self, **kwargs):  # pylint:disable=useless-super-delegation
        """Class representing the *Sazonov Algorithm* for sleep/wake detection based on activity counts.

        The *Sazonov Algorithm* runs sleep wake detection on epoch level activity data. Epochs are 1 minute long and
        activity is represented by an activity index which comes from Actigraph data or from raw acceleration data
        converted into activity index data.

        References
        ----------
        add reference

        """
        super().__init__(**kwargs)

    def fit(self, data: arr_t, **kwargs):  # noqa: ARG002
        """Fit sleep/wake detection algorithm to input data.

        .. note::
            Algorithms that do not have to (re)fit a ML model before sleep/wake prediction, such as rule-based
            algorithms, will internally bypass this method as the ``fit`` step is not needed.

        Parameters
        ----------
        data : array_like
            input data

        """
        return

    def predict(self, data: arr_t, **kwargs) -> SleepWakeDataFrame:
        """Apply sleep/wake prediction algorithm on input data.

        Parameters
        ----------
        data : array_like
            input data with activity index values
        **kwargs :
            additional arguments to be passed to the algorithm for prediction, such as:

            * ``rescore_data`` (``bool``):
              ``True`` to apply Webster's rescoring rules to the sleep/wake predictions, ``False`` otherwise.
              Default: ``True``
            * ``epoch_length`` (``int``):
              activity data epoch lengths in seconds, i.e. Epoch lengths are usually 30 or 60 seconds.
              Default: 30

        Returns
        -------
        :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`
            dataframe with sleep/wake predictions

        """
        index = None
        rescore_data: bool = kwargs.get("rescore_data", True)
        epoch_length: bool = kwargs.get("epoch_length", 30)

        if isinstance(data, pd.DataFrame):
            index = data.index
            data = sanitize_input_1d(data)

        w0 = data
        w1 = np.pad(np.max(sliding_window(data, window_samples=2, overlap_samples=1), axis=1), (1, 0))
        w2 = np.pad(np.max(sliding_window(data, window_samples=3, overlap_samples=2), axis=1), (2, 0))
        w3 = np.pad(np.max(sliding_window(data, window_samples=4, overlap_samples=3), axis=1), (3, 0))
        w4 = np.pad(np.max(sliding_window(data, window_samples=5, overlap_samples=4), axis=1), (4, 0))

        scores = np.array(1.727 - 0.256 * w0 - 0.154 * w1 - 0.136 * w2 - 0.140 * w3 - 0.176 * w4)
        scores = expit(scores)

        scores[scores >= 0.5] = 1
        scores[scores < 0.5] = 0

        if rescore_data:
            scores = rescore(scores, epoch_length=epoch_length)

        if index is not None:
            scores = pd.DataFrame(scores, index=index, columns=["sleep_wake"])

        return _SleepWakeDataFrame(scores)
